# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib as mpl
import matplotlib.pyplot as plt
import pytest
from numpy import array

from lightworks import (
    PostSelection,
    PostSelectionFunction,
    ResultCreationError,
    State,
    convert,
)
from lightworks.sdk.results import (
    SamplingResult,
    SimulationResult,
)

CUSTOM_PS = PostSelection()
CUSTOM_PS.add(1, (1, 3))

CUSTOM_PS2 = PostSelection()
CUSTOM_PS2.add(2, (1, 3))


class TestSamplingResult:
    """Unit tests for SamplingResult object."""

    def setup_class(self) -> None:
        """Create a variety of useful pieces of data for testing."""
        self.test_input = State([1, 1, 0, 0])
        self.test_dict = {
            State([1, 0, 0, 1]): 0.3,
            State([0, 1, 0, 1]): 0.4,
            State([0, 0, 2, 0]): 0.3,
            State([0, 3, 0, 1]): 0.2,
        }

    def test_dict_result_creation(self):
        """
        Checks that a result object can be created with a dictionary
        successfully.
        """
        SamplingResult(self.test_dict, self.test_input)

    def test_invalid_input_type(self):
        """
        Checks that input state is required to be a State object.
        """
        with pytest.raises(ResultCreationError):
            SamplingResult(self.test_dict, self.test_input.s)

    def test_single_input_retrival(self):
        """
        Confirms that result retrieval works correctly for single input case.
        """
        r = SamplingResult(self.test_dict, self.test_input)
        assert r[State([0, 1, 0, 1])] == 0.4

    def test_items(self):
        """Test return value from items method is correct."""
        r = SamplingResult(self.test_dict, self.test_input)
        assert r.items() == self.test_dict.items()

    def test_keys(self):
        """
        Checks that keys attribute returns a list of the output states.
        """
        r = SamplingResult(self.test_dict, self.test_input)
        assert list(r.keys()) == r.outputs

    @pytest.mark.parametrize("invert", [True, False])
    def test_threshold_mapping(self, invert):
        """Check threshold mapping returns the correct result."""
        r = SamplingResult(self.test_dict, self.test_input)
        r2 = r.map(convert.threshold_mapping, invert=invert)
        # Round results returned from mapping and compare
        out_dict = {s: round(p, 4) for s, p in r2.items()}
        if not invert:
            expected = {
                State([1, 0, 0, 1]): 0.3,
                State([0, 1, 0, 1]): 0.6,
                State([0, 0, 1, 0]): 0.3,
            }
        else:
            expected = {
                State([0, 1, 1, 0]): 0.3,
                State([1, 0, 1, 0]): 0.6,
                State([1, 1, 0, 1]): 0.3,
            }
        assert out_dict == expected

    @pytest.mark.parametrize("invert", [True, False])
    def test_parity_mapping(self, invert):
        """Check parity mapping returns the correct result."""
        r = SamplingResult(self.test_dict, self.test_input)
        r2 = r.map(convert.parity_mapping, invert=invert)
        # Round results returned from mapping and compare
        out_dict = {s: round(p, 4) for s, p in r2.items()}
        if not invert:
            expected = {
                State([1, 0, 0, 1]): 0.3,
                State([0, 1, 0, 1]): 0.6,
                State([0, 0, 0, 0]): 0.3,
            }
        else:
            expected = {
                State([0, 1, 1, 0]): 0.3,
                State([1, 0, 1, 0]): 0.6,
                State([1, 1, 1, 1]): 0.3,
            }
        assert out_dict == expected

    def test_custom_mapping(self):
        """
        Checks custom mapping can be applied correctly to create a new result.
        """

        def mapping(state: State) -> State:  # noqa: FURB118
            return state[2:]

        r = SamplingResult(self.test_dict, self.test_input)
        # Apply mapping
        r2 = r.map(mapping)
        # Round results returned from mapping and compare
        out_dict = {s: round(p, 4) for s, p in r2.items()}
        assert out_dict == {State([0, 1]): 0.9, State([2, 0]): 0.3}

    def test_custom_mapping_kwargs(self):
        """
        Checks that the arguments are correctly passed to a custom mapping by
        getting it to raise an exception.
        """

        class CustomTestError(Exception): ...

        def mapping(state: State, test: bool) -> State:
            if test:
                raise CustomTestError("Raised!")
            return state[2:]

        r = SamplingResult(self.test_dict, self.test_input)
        # Check mapping can be applied normally
        r.map(mapping, False)
        # Check exception raised when passed as arg
        with pytest.raises(CustomTestError):
            r.map(mapping, True)
        # And then as kwarg
        with pytest.raises(CustomTestError):
            r.map(mapping, test=True)

    @pytest.mark.parametrize(
        "post_selection",
        [
            CUSTOM_PS,
            PostSelectionFunction(lambda s: s[1] > 0),
            lambda s: s[1] > 0,
        ],
    )
    def test_custom_post_selection(self, post_selection):
        """
        Check that additional post-selection can be applied to a result
        correctly.
        """
        r = SamplingResult(self.test_dict, self.test_input)
        r2 = r.apply_post_selection(post_selection)
        # Compare results
        assert r2 == {State([0, 1, 0, 1]): 0.4, State([0, 3, 0, 1]): 0.2}

    def test_single_input_plot(self):
        """
        Confirm plotting is able to work without errors for single input case.
        """
        r = SamplingResult(self.test_dict, self.test_input)
        # NOTE: There is a intermittent issue that occurs during testing
        # with the subplots method in mpl. This can be fixed by altering the
        # backend to Agg for these tests. Issue noted here:
        # https://stackoverflow.com/questions/71443540/intermittent-pytest-failures-complaining-about-missing-tcl-files-even-though-the
        original_backend = mpl.get_backend()
        mpl.use("Agg")
        # Check plotting
        r.plot(show=False)
        plt.close()
        # Reset backend after test
        mpl.use(original_backend)

    def test_extra_attribute_assignment(self):
        """
        Check that miscellaneous additional kwargs can be provided in the
        initial function call and that these are assigned to attributes.
        """
        r = SamplingResult(self.test_dict, self.test_input, test_attr=2.5)
        assert r.test_attr == 2.5

    def test_print_outputs(self):
        """
        Checks no exceptions are raised when the print outputs method is
        called.
        """
        r = SamplingResult(self.test_dict, self.test_input)
        r.print_outputs()

    def test_display_as_dataframe(self):
        """
        Checks that no exceptions are raised when display as Dataframe method
        is called.
        """
        r = SamplingResult(self.test_dict, self.test_input)
        r.display_as_dataframe()

    def test_iterable(self):
        """
        Checks that SamplingResult acts an iterable which will run through all
        outputs included in the result.
        """
        r = SamplingResult(self.test_dict, self.test_input)
        values = []
        for i in r:
            values.append(i)
        assert values == r.outputs

    def test_str(self):
        """
        Tests that string return should be a string representation of the
        results dictionary
        """
        r = SamplingResult(self.test_dict, self.test_input)
        assert str(r) == str(self.test_dict)

    def test_get_items(self):
        """
        Checks that get item returns correct value for output.
        """
        r = SamplingResult(self.test_dict, self.test_input)
        assert r[State([1, 0, 0, 1])] == 0.3

    def test_get_item_invalid_state(self):
        """
        Checks that get item raises a KeyError when output isn't found.
        """
        r = SamplingResult(self.test_dict, self.test_input)
        with pytest.raises(KeyError):
            r[State([1, 0, 2, 1])]

    @pytest.mark.parametrize("value", [[0, 1, 2, 3], [0], ["|1,0,0,1>"]])
    def test_get_item_invalid_type(self, value):
        """
        Checks that get item returns a TypeError when a non-state key is used.
        """
        r = SamplingResult(self.test_dict, self.test_input)
        with pytest.raises(TypeError):
            r[value]


class TestSimulationResult:
    """Unit tests for SimulationResult object."""

    def setup_method(self) -> None:
        """Create a variety of useful pieces of data for testing."""
        # Single input
        self.test_single_inputs = [State([1, 1, 0, 0])]
        self.test_single_outputs = [
            State([1, 0, 1, 0]),
            State([0, 1, 0, 1]),
            State([1, 0, 3, 0]),
        ]
        self.test_single_array = array([[0.3, 0.2, 0.1]])
        # Multiple inputs
        self.test_multi_inputs = [State([1, 1, 0, 0]), State([0, 0, 1, 1])]
        self.test_multi_outputs = [
            State([1, 0, 1, 0]),
            State([0, 1, 0, 1]),
            State([1, 0, 3, 0]),
        ]
        self.test_multi_array = array([[0.3, 0.2, 0.1], [0.2, 0.4, 0.5]])

    def test_single_array_result_creation(self):
        """
        Checks that a result object can be created with an array successfully
        with a single input.
        """
        SimulationResult(
            self.test_single_array,
            "probability_amplitude",
            inputs=self.test_single_inputs,
            outputs=self.test_single_outputs,
        )

    def test_multi_array_result_creation(self):
        """
        Checks that a result object can be created with an array successfully
        with multiple inputs.
        """
        SimulationResult(
            self.test_multi_array,
            "probability_amplitude",
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
        )

    def test_single_input_retrival(self):
        """
        Checks that output can be indexed in case of single input.
        """
        r = SimulationResult(
            self.test_single_array,
            "probability_amplitude",
            inputs=self.test_single_inputs,
            outputs=self.test_single_outputs,
        )
        assert r[self.test_single_inputs[0]][State([1, 0, 1, 0])] == 0.3

    def test_multi_input_retrival(self):
        """
        Confirms that result retrieval works correctly for multi input case.
        """
        r = SimulationResult(
            self.test_multi_array,
            "probability_amplitude",
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
        )
        assert r[State([1, 1, 0, 0])] == {
            State([1, 0, 1, 0]): 0.3,
            State([0, 1, 0, 1]): 0.2,
            State([1, 0, 3, 0]): 0.1,
        }

    def test_result_indexing(self):
        """
        Confirms that result retrieval works correctly for multi input case,
        with both the input and output used to get a single value.
        """
        r = SimulationResult(
            self.test_multi_array,
            "probability_amplitude",
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
        )
        assert r[State([1, 1, 0, 0]), State([1, 0, 3, 0])] == 0.1

    @pytest.mark.parametrize(
        ("conv_to_probability", "rtype"),
        [
            (True, "probability_amplitude"),
            (False, "probability_amplitude"),
            (True, "probability"),
            (False, "probability"),
        ],
    )
    def test_multi_input_plot(self, conv_to_probability, rtype):
        """
        Confirm plotting is able to work without errors for multi input case.
        """
        # NOTE: There is a intermittent issue that occurs during testing
        # with the subplots method in mpl. This can be fixed by altering the
        # backend to Agg for these tests. Issue noted here:
        # https://stackoverflow.com/questions/71443540/intermittent-pytest-failures-complaining-about-missing-tcl-files-even-though-the
        original_backend = mpl.get_backend()
        mpl.use("Agg")
        r = SimulationResult(
            self.test_multi_array,
            rtype,
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
        )
        # Test plot
        r.plot(conv_to_probability=conv_to_probability, show=False)
        plt.close()
        # Reset backend after test
        mpl.use(original_backend)

    @pytest.mark.parametrize(
        ("conv_to_probability", "rtype"),
        [
            (True, "probability_amplitude"),
            (False, "probability_amplitude"),
            (True, "probability"),
            (False, "probability"),
        ],
    )
    def test_single_input_plot(self, conv_to_probability, rtype):
        """
        Confirm plotting is able to work without errors for single input case.
        """
        original_backend = mpl.get_backend()
        mpl.use("Agg")
        r = SimulationResult(
            self.test_single_array,
            rtype,
            inputs=self.test_single_inputs,
            outputs=self.test_single_outputs,
        )
        # Test plot
        r.plot(conv_to_probability=conv_to_probability, show=False)
        plt.close()
        # Reset backend after test
        mpl.use(original_backend)

    def test_extra_attribute_assignment(self):
        """
        Check that miscellaneous additional kwargs can be provided in the
        initial function call and that these are assigned to attributes.
        """
        r = SimulationResult(
            self.test_multi_array,
            "probability_amplitude",
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
            performance=2.5,
        )
        assert r.performance == 2.5

    def test_single_print_outputs(self):
        """
        Confirms print outputs runs without raising an exception.
        """
        r = SimulationResult(
            self.test_single_array,
            "probability_amplitude",
            inputs=self.test_single_inputs,
            outputs=self.test_single_outputs,
        )
        r.print_outputs()

    def test_multi_print_outputs(self):
        """
        Confirms print outputs runs without raising an exception.
        """
        r = SimulationResult(
            self.test_multi_array,
            "probability_amplitude",
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
        )
        r.print_outputs()

    @pytest.mark.parametrize("conv_to_probability", [True, False])
    def test_single_display_as_dataframe(self, conv_to_probability):
        """
        Confirms display as dataframe runs without raising an exception.
        """
        r = SimulationResult(
            self.test_single_array,
            "probability_amplitude",
            inputs=self.test_single_inputs,
            outputs=self.test_single_outputs,
        )
        r.display_as_dataframe(conv_to_probability=conv_to_probability)

    @pytest.mark.parametrize("conv_to_probability", [True, False])
    def test_multi_display_as_dataframe(self, conv_to_probability):
        """
        Confirms display as dataframe runs without raising an exception.
        """
        r = SimulationResult(
            self.test_multi_array,
            "probability_amplitude",
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
        )
        r.display_as_dataframe(conv_to_probability=conv_to_probability)

    def test_single_input_parity_mapping(self):
        """
        Tests result is correct when the parity mapping is applied in the
        single input case.
        """
        r = SimulationResult(
            self.test_single_array,
            "probability",
            inputs=self.test_single_inputs,
            outputs=self.test_single_outputs,
        )
        new_r = r.map(convert.parity_mapping)
        assert new_r == {
            State([1, 1, 0, 0]): {
                State([1, 0, 1, 0]): 0.4,
                State([0, 1, 0, 1]): 0.2,
            }
        }

    @pytest.mark.parametrize("invert", [True, False])
    def test_multi_input_parity_mapping(self, invert):
        """
        Tests result is correct when the parity mapping is applied in the
        multi input case.
        """
        r = SimulationResult(
            self.test_multi_array,
            "probability",
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
        )
        new_r = r.map(convert.parity_mapping, invert=invert)
        if not invert:
            expected = {
                State([1, 1, 0, 0]): {
                    State([1, 0, 1, 0]): 0.4,
                    State([0, 1, 0, 1]): 0.2,
                },
                State([0, 0, 1, 1]): {
                    State([1, 0, 1, 0]): 0.7,
                    State([0, 1, 0, 1]): 0.4,
                },
            }
        else:
            expected = {
                State([1, 1, 0, 0]): {
                    State([1, 0, 1, 0]): 0.2,
                    State([0, 1, 0, 1]): 0.4,
                },
                State([0, 0, 1, 1]): {
                    State([1, 0, 1, 0]): 0.4,
                    State([0, 1, 0, 1]): 0.7,
                },
            }
        assert new_r == expected

    def test_single_input_threshold_mapping(self):
        """
        Tests result is correct when the threshold mapping is applied in the
        single input case.
        """
        r = SimulationResult(
            self.test_single_array,
            "probability",
            inputs=self.test_single_inputs,
            outputs=self.test_single_outputs,
        )
        new_r = r.map(convert.threshold_mapping)
        assert new_r == {
            State([1, 1, 0, 0]): {
                State([1, 0, 1, 0]): 0.4,
                State([0, 1, 0, 1]): 0.2,
            }
        }

    @pytest.mark.parametrize("invert", [True, False])
    def test_multi_input_threshold_mapping(self, invert):
        """
        Tests result is correct when the threshold mapping is applied in the
        single multi case.
        """
        r = SimulationResult(
            self.test_multi_array,
            "probability",
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
        )
        new_r = r.map(convert.threshold_mapping, invert=invert)
        if not invert:
            expected = {
                State([1, 1, 0, 0]): {
                    State([1, 0, 1, 0]): 0.4,
                    State([0, 1, 0, 1]): 0.2,
                },
                State([0, 0, 1, 1]): {
                    State([1, 0, 1, 0]): 0.7,
                    State([0, 1, 0, 1]): 0.4,
                },
            }
        else:
            expected = {
                State([1, 1, 0, 0]): {
                    State([1, 0, 1, 0]): 0.2,
                    State([0, 1, 0, 1]): 0.4,
                },
                State([0, 0, 1, 1]): {
                    State([1, 0, 1, 0]): 0.4,
                    State([0, 1, 0, 1]): 0.7,
                },
            }
        assert new_r == expected

    def test_custom_mapping(self):
        """
        Checks custom mapping can be applied correctly to create a new result.
        """

        def mapping(state: State) -> State:  # noqa: FURB118
            return state[:2]

        r = SimulationResult(
            self.test_multi_array,
            "probability",
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
        )
        # Apply mapping
        r2 = r.map(mapping)
        # Compare results
        assert r2 == {
            State([1, 1, 0, 0]): {State([1, 0]): 0.4, State([0, 1]): 0.2},
            State([0, 0, 1, 1]): {State([1, 0]): 0.7, State([0, 1]): 0.4},
        }

    def test_custom_mapping_kwargs(self):
        """
        Checks that the arguments are correctly passed to a custom mapping by
        getting it to raise an exception.
        """

        class CustomTestError(Exception): ...

        def mapping(state: State, test: bool) -> State:
            if test:
                raise CustomTestError("Raised!")
            return state[:2]

        r = SimulationResult(
            self.test_multi_array,
            "probability",
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
        )
        # Check mapping can be applied normally
        r.map(mapping, False)
        # Check exception raised when passed as arg
        with pytest.raises(CustomTestError):
            r.map(mapping, True)
        # And then as kwarg
        with pytest.raises(CustomTestError):
            r.map(mapping, test=True)

    @pytest.mark.parametrize(
        "post_selection",
        [
            CUSTOM_PS2,
            PostSelectionFunction(lambda s: s[2] > 0),
            lambda s: s[2] > 0,
        ],
    )
    def test_custom_post_selection(self, post_selection):
        """
        Check that additional post-selection can be applied to a result
        correctly.
        """
        r = SimulationResult(
            self.test_multi_array,
            "probability",
            inputs=self.test_multi_inputs,
            outputs=self.test_multi_outputs,
        )
        r2 = r.apply_post_selection(post_selection)
        # Compare results
        assert r2 == {
            State([1, 1, 0, 0]): {
                State([1, 0, 1, 0]): 0.3,
                State([1, 0, 3, 0]): 0.1,
            },
            State([0, 0, 1, 1]): {
                State([1, 0, 1, 0]): 0.2,
                State([1, 0, 3, 0]): 0.5,
            },
        }



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains all custom exceptions created for the emulator section of Lightworks.
"""

from lightworks.sdk.utils.exceptions import LightworksError


class EmulatorError(LightworksError):
    """
    Generic error for emulator which sub classes the main LightworksError
    """


class AnnotatedStateError(EmulatorError):
    """
    Error relating to issues with a provided AnnotatedState
    """


class BackendError(EmulatorError):
    """
    Raised when errors occur in the Backend object.
    """



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A custom state datatype, which is created with the aim of making states in the
emulator much easier to work with.
"""

from collections.abc import Iterator
from copy import copy
from typing import Any, Union, overload

from lightworks.sdk.utils.exceptions import StateError


class State:
    """
    Custom data type to store information about a quantum state, as well as
    allowing a number of operations to act on the state.

    Args:

        state (list) : The fock basis state to use with the class, this should
            be a list of photon numbers per mode.

    """

    __slots__ = ["__s"]

    def __init__(self, state: list[int]) -> None:
        # If already list then assign to attribute
        if isinstance(state, list):
            self.__s = state
        # Otherwise try to convert
        else:
            self.__s = list(state)

    @property
    def n_photons(self) -> int:
        """Returns the number of photons in a State."""
        return sum(self.__s)

    @property
    def s(self) -> list[int]:
        """Returns a copy of the contents of the state as a list."""
        return copy(self.__s)

    @s.setter
    def s(self, value: Any) -> None:  # noqa: ARG002
        raise StateError("State value should not be modified directly.")

    @property
    def n_modes(self) -> int:
        """The total number of modes in the state."""
        return len(self.__s)

    @n_modes.setter
    def n_modes(self, value: Any) -> None:  # noqa: ARG002
        raise StateError("Number of modes cannot be modified.")

    def merge(self, merge_state: "State") -> "State":
        """Combine two states, summing the number of photons per mode."""
        if self.n_modes != merge_state.n_modes:
            raise ValueError("Merged states must be the same length.")
        return State(
            [n1 + n2 for n1, n2 in zip(self.__s, merge_state.s, strict=True)]
        )

    def _validate(self) -> None:
        """
        Function to perform some validation of a state, including checking that
        the values are all integers and that no negative values are included.
        """
        for s in self.__s:
            if not isinstance(s, int) or isinstance(s, bool):
                raise TypeError(
                    "State mode occupation numbers should be integers."
                )
            if s < 0:
                raise ValueError("Mode occupation numbers cannot be negative.")

    def __str__(self) -> str:
        return state_to_string(self.__s)

    def __repr__(self) -> str:
        return f"lightworks.State({state_to_string(self.__s)})"

    def __add__(self, value: "State") -> "State":
        if not isinstance(value, State):
            raise TypeError("Addition only supported between states.")
        return State(self.__s + value.__s)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, State):
            return False
        return self.__s == value.s

    def __hash__(self) -> int:
        return hash(self.__str__())

    def __len__(self) -> int:
        return self.n_modes

    def __iter__(self) -> Iterator[int]:
        yield from self.s

    def __setitem__(self, key: Any, value: Any) -> None:
        raise StateError("State object does not support item assignment.")

    @overload
    def __getitem__(self, indices: int) -> int: ...

    @overload
    def __getitem__(self, indices: slice) -> "State": ...

    def __getitem__(self, indices: slice | int) -> Union[int, "State"]:
        if isinstance(indices, slice):
            return State(self.__s[indices])
        if isinstance(indices, int):
            return self.__s[indices]
        raise TypeError("Subscript should either be int or slice.")


def state_to_string(state: list[int]) -> str:
    """Converts the provided state to a string with ket notation."""
    string = "|"
    for s in state:
        string += str(s) + ","
    return string[:-1] + ">"



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ruff: noqa: N802, I001

import numpy as np
import pytest
from random import random

from lightworks import State, Simulator, Parameter
from lightworks.emulator import Backend

# fmt: off
from lightworks.qubit import (
    CCNOT, CCZ, CNOT, CZ, SWAP, CNOT_Heralded, CZ_Heralded, I, H, S, T, X, Y, Z,
    SX, Sadj, Tadj, P, Rx, Ry, Rz
)
import math
# fmt: on

BACKEND = Backend("permanent")


class TestSingleQubitGates:
    """
    Unit tests for all single qubit gates.
    """

    def test_I(self):
        """Checks that the output from the I gate is correct."""
        sim = Simulator(I(), State([1, 0]))
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 1
        assert pytest.approx(results[State([0, 1])], 1e-6) == 0
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 0
        assert pytest.approx(results[State([0, 1])], 1e-6) == 1

    def test_hadamard(self):
        """Checks that the output from the Hadamard gate is correct."""
        sim = Simulator(H(), State([1, 0]))
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 2**-0.5
        assert pytest.approx(results[State([0, 1])], 1e-6) == 2**-0.5
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 2**-0.5
        assert pytest.approx(results[State([0, 1])], 1e-6) == -(2**-0.5)

    def test_X(self):
        """Checks that the output from the X gate is correct."""
        sim = Simulator(X(), State([1, 0]))
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 0
        assert pytest.approx(results[State([0, 1])], 1e-6) == 1
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 1
        assert pytest.approx(results[State([0, 1])], 1e-6) == 0

    @pytest.mark.parametrize("theta", [math.pi, Parameter(math.pi)])
    def test_Rx(self, theta):
        """Checks that the output from the Rx gate is correct."""
        gate = Rx(theta)
        gate.ps(0, math.pi / 2)  # Make equivalent to X gate
        gate.ps(1, math.pi / 2)
        sim = Simulator(gate, State([1, 0]))
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 0
        assert pytest.approx(results[State([0, 1])], 1e-6) == 1
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 1
        assert pytest.approx(results[State([0, 1])], 1e-6) == 0

    def test_Y(self):
        """Checks that the output from the Y gate is correct."""
        sim = Simulator(Y(), State([1, 0]))
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 0
        assert pytest.approx(results[State([0, 1])], 1e-6) == 1j
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == -1j
        assert pytest.approx(results[State([0, 1])], 1e-6) == 0

    @pytest.mark.parametrize("theta", [math.pi, Parameter(math.pi)])
    def test_Ry(self, theta):
        """Checks that the output from the Ry gate is correct."""
        gate = Ry(theta)
        gate.ps(0, math.pi / 2)  # Make equivalent to Y gate
        gate.ps(1, math.pi / 2)
        sim = Simulator(gate, State([1, 0]))
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 0
        assert pytest.approx(results[State([0, 1])], 1e-6) == 1j
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == -1j
        assert pytest.approx(results[State([0, 1])], 1e-6) == 0

    def test_Z(self):
        """Checks that the output from the Z gate is correct."""
        sim = Simulator(Z(), State([1, 0]))
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 1
        assert pytest.approx(results[State([0, 1])], 1e-6) == 0
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 0
        assert pytest.approx(results[State([0, 1])], 1e-6) == -1

    @pytest.mark.parametrize("theta", [math.pi, Parameter(math.pi)])
    def test_Rz(self, theta):
        """Checks that the output from the Rz gate is correct."""
        gate = Rz(theta)
        gate.ps(0, math.pi / 2)  # Make equivalent to Z gate
        gate.ps(1, math.pi / 2)
        sim = Simulator(gate, State([1, 0]))
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 1
        assert pytest.approx(results[State([0, 1])], 1e-6) == 0
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 0
        assert pytest.approx(results[State([0, 1])], 1e-6) == -1

    def test_S(self):
        """Checks that the output from the S gate is correct."""
        sim = Simulator(S(), State([1, 0]))
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 1
        assert pytest.approx(results[State([0, 1])], 1e-6) == 0
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 0
        assert pytest.approx(results[State([0, 1])], 1e-6) == 1j

    def test_Sadj(self):
        """Checks that Sadj gate is hermitian conjugate of S."""
        assert pytest.approx(S().U, 1e-8) == np.conj(Sadj().U).T

    def test_T(self):
        """Checks that the output from the T gate is correct."""
        sim = Simulator(T(), State([1, 0]))
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 1
        assert pytest.approx(results[State([0, 1])], 1e-6) == 0
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 0
        assert pytest.approx(results[State([0, 1])], 1e-6) == np.exp(
            1j * np.pi / 4
        )

    def test_Tadj(self):
        """Checks that Tadj gate is hermitian conjugate of S."""
        assert pytest.approx(T().U, 1e-8) == np.conj(Tadj().U).T

    def test_SX(self):
        """Checks that the output from the SX gate is correct."""
        sim = Simulator(SX(), State([1, 0]))
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 1 / 2 * (1 + 1j)
        assert pytest.approx(results[State([0, 1])], 1e-6) == 1 / 2 * (1 - 1j)
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 1 / 2 * (1 - 1j)
        assert pytest.approx(results[State([0, 1])], 1e-6) == 1 / 2 * (1 + 1j)

    @pytest.mark.parametrize("phase", [None, Parameter(math.tau * random())])
    def test_P(self, phase):
        """Checks that the output from the phase gate is correct."""
        if phase is None:
            phase = math.tau * random()
        sim = Simulator(P(phase), State([1, 0]))
        if isinstance(phase, Parameter):
            phase = phase.get()
        # Input |1,0>
        results = BACKEND.run(sim)[State([1, 0])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 1
        assert pytest.approx(results[State([0, 1])], 1e-6) == 0
        # Input |0,1>
        sim.inputs = State([0, 1])
        results = BACKEND.run(sim)[State([0, 1])]
        assert pytest.approx(results[State([1, 0])], 1e-6) == 0
        assert pytest.approx(results[State([0, 1])], 1e-6) == np.exp(1j * phase)


class TestTwoQubitGates:
    """
    Unit tests for all single two gates.
    """

    def test_CZ(self):
        """
        Checks that the output of the post-selected CZ gate is correct and that
        the success probability is 1/9.
        """
        # Define all input combinations
        states = [[1, 0, 1, 0], [1, 0, 0, 1], [0, 1, 1, 0], [0, 1, 0, 1]]
        states = [State(s) for s in states]
        # Calculate probability amplitudes
        sim = Simulator(CZ(), states, states)
        results = BACKEND.run(sim)
        # Check all results are identical except for |1,1> which has a -1
        amp = results[states[0], states[0]]
        assert pytest.approx(amp, 1e-6) == results[states[1], states[1]]
        assert pytest.approx(amp, 1e-6) == results[states[2], states[2]]
        assert pytest.approx(amp, 1e-6) == -results[states[3], states[3]]
        # Confirm success probability is 1/9
        assert pytest.approx(abs(amp) ** 2, 1e-6) == 1 / 9

    def test_CNOT(self):
        """
        Checks that the output of the post-selected CNOT gate is correct and
        that the success probability is 1/9.
        """
        # Define all input combinations
        states = [[1, 0, 1, 0], [1, 0, 0, 1], [0, 1, 1, 0], [0, 1, 0, 1]]
        states = [State(s) for s in states]
        # Calculate probability amplitudes
        sim = Simulator(CNOT(), states, states)
        results = BACKEND.run(sim)
        # Check that swap occurs when control qubit is 1 but not otherwise
        amp = results[states[0], states[0]]
        assert pytest.approx(amp, 1e-6) == results[states[1], states[1]]
        assert pytest.approx(amp, 1e-6) == results[states[2], states[3]]
        assert pytest.approx(amp, 1e-6) == results[states[3], states[2]]
        # Confirm success probability is 1/9
        assert pytest.approx(abs(amp) ** 2, 1e-6) == 1 / 9

    def test_CNOT_flipped(self):
        """
        Checks that the output of the post-selected CNOT gate is correct and
        that the success probability is 1/9 when the target qubit is set to 0.
        """
        # Define all input combinations
        states = [[1, 0, 1, 0], [1, 0, 0, 1], [0, 1, 1, 0], [0, 1, 0, 1]]
        states = [State(s) for s in states]
        # Calculate probability amplitudes
        sim = Simulator(CNOT(target_qubit=0), states, states)
        results = BACKEND.run(sim)
        # Check that swap occurs when control qubit is 1 but not otherwise
        amp = results[states[0], states[0]]
        assert pytest.approx(amp, 1e-6) == results[states[2], states[2]]
        assert pytest.approx(amp, 1e-6) == results[states[1], states[3]]
        assert pytest.approx(amp, 1e-6) == results[states[3], states[1]]
        # Confirm success probability is 1/9
        assert pytest.approx(abs(amp) ** 2, 1e-6) == 1 / 9

    def test_CZ_heralded(self):
        """
        Checks that the output of the heralded CZ gate is correct and that the
        success probability is 1/16.
        """
        # Define all input combinations
        states = [[1, 0, 1, 0], [1, 0, 0, 1], [0, 1, 1, 0], [0, 1, 0, 1]]
        states = [State(s) for s in states]
        # Calculate probability amplitudes
        sim = Simulator(CZ_Heralded(), states, states)
        results = BACKEND.run(sim)
        # Check all results are identical except for |1,1> which has a -1
        amp = results[states[0], states[0]]
        assert pytest.approx(amp, 1e-6) == results[states[1], states[1]]
        assert pytest.approx(amp, 1e-6) == results[states[2], states[2]]
        assert pytest.approx(amp, 1e-6) == -results[states[3], states[3]]
        # Confirm success probability is 1/16
        assert pytest.approx(abs(amp) ** 2, 1e-6) == 1 / 16

    def test_CNOT_heralded(self):
        """
        Checks that the output of the heralded CNOT gate is correct and that
        the success probability is 1/16.
        """
        # Define all input combinations
        states = [[1, 0, 1, 0], [1, 0, 0, 1], [0, 1, 1, 0], [0, 1, 0, 1]]
        states = [State(s) for s in states]
        # Calculate probability amplitudes
        sim = Simulator(CNOT_Heralded(), states, states)
        results = BACKEND.run(sim)
        # Check that swap occurs when control qubit is 1 but not otherwise
        amp = results[states[0], states[0]]
        assert pytest.approx(amp, 1e-6) == results[states[1], states[1]]
        assert pytest.approx(amp, 1e-6) == results[states[2], states[3]]
        assert pytest.approx(amp, 1e-6) == results[states[3], states[2]]
        # Confirm success probability is 1/16
        assert pytest.approx(abs(amp) ** 2, 1e-6) == 1 / 16

    def test_CNOT_heralded_flipped(self):
        """
        Checks that the output of the heralded CNOT gate is correct and that
        the success probability is 1/16 when the target qubit is set to 0.
        """
        # Define all input combinations
        states = [[1, 0, 1, 0], [1, 0, 0, 1], [0, 1, 1, 0], [0, 1, 0, 1]]
        states = [State(s) for s in states]
        # Calculate probability amplitudes
        sim = Simulator(CNOT_Heralded(target_qubit=0), states, states)
        results = BACKEND.run(sim)
        # Check that swap occurs when control qubit is 1 but not otherwise
        amp = results[states[0], states[0]]
        assert pytest.approx(amp, 1e-6) == results[states[2], states[2]]
        assert pytest.approx(amp, 1e-6) == results[states[1], states[3]]
        assert pytest.approx(amp, 1e-6) == results[states[3], states[1]]
        # Confirm success probability is 1/16
        assert pytest.approx(abs(amp) ** 2, 1e-6) == 1 / 16

    def test_swap(self):
        """
        Checks that SWAP gate correctly maps modes to the correct location.
        """
        # Define gate
        swap = SWAP((0, 1), (4, 5))
        # Create simulator
        sim = Simulator(
            swap, State([1, 0, 0, 0, 0, 0]), State([0, 0, 0, 0, 1, 0])
        )
        # Check 0 mode of first qubit maps to 0 of second qubit
        results = BACKEND.run(sim)
        amp = results.array[0, 0]
        assert amp == pytest.approx(1 + 0j, 1e-6)
        # Check 1 mode of second qubit maps to 1 of first qubit
        sim.inputs = State([0, 0, 0, 0, 0, 1])
        sim.outputs = State([0, 1, 0, 0, 0, 0])
        results = BACKEND.run(sim)
        amp = results.array[0, 0]
        assert amp == pytest.approx(1 + 0j, 1e-6)


class TestThreeQubitGates:
    """
    Unit tests for all three qubit gates.
    """

    def test_CCZ(self):
        """
        Checks that the output of the post-selected CCZ gate is correct and
        that the success probability is 1/72.
        """
        # Define all input combinations
        states = [
            [1, 0, 1, 0, 1, 0],
            [1, 0, 0, 1, 1, 0],
            [0, 1, 1, 0, 1, 0],
            [0, 1, 0, 1, 1, 0],
            [1, 0, 1, 0, 0, 1],
            [1, 0, 0, 1, 0, 1],
            [0, 1, 1, 0, 0, 1],
            [0, 1, 0, 1, 0, 1],
        ]
        states = [State(s) for s in states]
        # Calculate probability amplitudes
        sim = Simulator(CCZ(), states, states)
        results = BACKEND.run(sim)
        # Check all results are identical except for |1,1,1> which has a -1
        amp = results[states[0], states[0]]
        for i in range(7):
            assert pytest.approx(amp, 1e-6) == results[states[i], states[i]]
        assert pytest.approx(amp, 1e-6) == -results[states[7], states[7]]
        # Confirm success probability is 1/72
        assert pytest.approx(abs(amp) ** 2, 1e-6) == 1 / 72

    @pytest.mark.parametrize(
        ("target", "swap_indices"), [(0, (5, 7)), (1, (6, 7)), (2, (3, 7))]
    )
    def test_CCNOT(self, target, swap_indices):
        """
        Checks that the output of the post-selected CCNOT gate is correct and
        that the success probability is 1/72. This is checked for all possible
        target qubits.
        """
        # Define all input combinations
        states = [
            [1, 0, 1, 0, 1, 0],
            [1, 0, 0, 1, 1, 0],
            [0, 1, 1, 0, 1, 0],
            [0, 1, 0, 1, 1, 0],
            [1, 0, 1, 0, 0, 1],
            [1, 0, 0, 1, 0, 1],
            [0, 1, 1, 0, 0, 1],
            [0, 1, 0, 1, 0, 1],
        ]
        states = [State(s) for s in states]
        # Calculate probability amplitudes
        sim = Simulator(CCNOT(target_qubit=target), states, states)
        results = BACKEND.run(sim)
        # Check swap occurs when both control qubits are 1 but not otherwise
        non_swapped = [i for i in range(8) if i not in swap_indices]
        amp = results[states[non_swapped[0]], states[non_swapped[0]]]
        for i in non_swapped:
            assert results[states[i], states[i]] == pytest.approx(amp, 1e-6)
        assert results[
            states[swap_indices[0]], states[swap_indices[1]]
        ] == pytest.approx(amp, 1e-6)
        assert results[
            states[swap_indices[1]], states[swap_indices[0]]
        ] == pytest.approx(amp, 1e-6)
        # Confirm success probability is 1/72
        assert pytest.approx(abs(amp) ** 2, 1e-6) == 1 / 72



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy

from lightworks.sdk.state import State


def add_heralds_to_state(
    state: State | list[int], heralds: dict[int, int]
) -> list[int]:
    """
    Takes a provided input state and includes any heralding photons/modes.

    Args:

        state (State | list) : The initial state with heralding modes excluded.

        heralds (dict) : A dictionary of the required heralds to include.

    Returns:

        list : The updated state with heralded modes included.

    """
    # Auto-return original state if no heralding used
    if not heralds:
        return state.s if isinstance(state, State) else copy(state)
    n_modes = len(state) + len(heralds)
    # Otherwise create new state
    new_state = [0] * n_modes
    # Then iterate through modes using values from state or herald
    count = 0
    for i in range(n_modes):
        if i in heralds:
            new_state[i] = heralds[i]
        else:
            new_state[i] = state[count]
            count += 1
    return new_state


def remove_heralds_from_state(
    state: State | list[int], herald_modes: list[int]
) -> list[int]:
    """
    Removes all heralded modes from a provided state.

    Args:

        state (State, list) : The state to remove heralds from.

        herald_modes (list) : A list of the modes used for heralding.

    Returns:

        list : The updated state with heralded modes removed.

    """
    # Remove modes in reverse order so mode locations do not change
    to_remove = sorted(herald_modes, reverse=True)
    # Get list version of state
    new_s = state.s if isinstance(state, State) else copy(state)
    # Then sequentially pop modes
    for m in to_remove:
        new_s.pop(m)
    return new_s



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

from lightworks.sdk.circuit import PhotonicCircuit
from lightworks.sdk.state import State
from lightworks.sdk.utils.post_selection import (
    PostSelectionType,
    process_post_selection,
)
from lightworks.sdk.utils.state import check_herald_difference, validate_states

from .data import AnalyzerTask
from .task import Task


class Analyzer(Task):
    """
    The analyzer class is built as an alternative to simulation, intended for
    cases where we want to look at the transformations between a specific
    subset of states. It is useful for the simulation of probabilities in
    cases where loss and circuit errors are likely to be a factor. As part of
    the process a performance and error rate metric are calculated.

    Args:

        circuit (PhotonicCircuit) : The circuit to simulate.

        inputs (list) : A list of the input states to simulate. For multiple
            inputs this should be a list of States.

        expected (dict) : A dictionary containing a mapping between the input
            state and expected output state(s). If there is multiple
            possible outputs, this can be specified as a list.

    Attribute:

        performance : The total probabilities of mapping between the states
            provided compared with all possible states.

        error_rate : Given an expected mapping, the analyzer will determine the
            extent to which this is achieved.

    """

    def __init__(
        self,
        circuit: PhotonicCircuit,
        inputs: State | list[State],
        expected: dict[State, State | list[State]] | None = None,
        post_selection: PostSelectionType
        | Callable[[State], bool]
        | None = None,
    ) -> None:
        # Assign key parameters to attributes
        self.circuit = circuit
        self.post_selection = post_selection  # type: ignore[assignment]
        self.inputs = inputs  # type: ignore[assignment]
        self.expected = expected

    @property
    def circuit(self) -> PhotonicCircuit:
        """
        Stores the circuit to be used for simulation, should be a
        PhotonicCircuit object.
        """
        return self.__circuit

    @circuit.setter
    def circuit(self, value: PhotonicCircuit) -> None:
        if not isinstance(value, PhotonicCircuit):
            raise TypeError(
                "Provided circuit should be a PhotonicCircuit or Unitary "
                "object."
            )
        self.__circuit = value

    @property
    def post_selection(self) -> PostSelectionType | None:
        """
        Stores post-selection criteria for analysis.
        """
        return self.__post_selection

    @post_selection.setter
    def post_selection(
        self, value: PostSelectionType | Callable[[State], bool] | None
    ) -> None:
        value = process_post_selection(value)
        self.__post_selection = value

    @property
    def inputs(self) -> list[State]:
        """Store list of target inputs to the system."""
        return self.__inputs

    @inputs.setter
    def inputs(self, value: State | list[State]) -> None:
        self.__inputs = validate_states(value, self.circuit.input_modes)

    @property
    def expected(self) -> dict[State, State | list[State]] | None:
        """
        A dictionary of the expected mapping between inputs and outputs of the
        system.
        """
        return self.__expected

    @expected.setter
    def expected(self, value: dict[State, State | list[State]] | None) -> None:
        self.__expected = value

    def _generate_task(self) -> AnalyzerTask:
        check_herald_difference(self.circuit, self.inputs[0].n_photons)
        return AnalyzerTask(
            circuit=self.circuit._build(),
            inputs=self.inputs,
            expected=self.expected,
            post_selection=self.post_selection,
        )



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lightworks.sdk.state import State
from lightworks.sdk.utils.exceptions import ProbabilityDistributionError


class ProbabilityDistribution(dict[State, float]):  # noqa: FURB189
    """
    Stores a created ProbabilityDistribution and prevents modification to
    values.
    """

    def __repr__(self) -> str:
        return f"lightworks.ProbabilityDistribution({super().__repr__()})"

    def __setitem__(self, key: State, value: complex) -> None:
        raise ProbabilityDistributionError(
            "Probability distribution should not be modified directly, to edit "
            "this make a copy using dict(ProbabilityDistribution)."
        )



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.typing import NDArray


def permutation_mat_from_swaps_dict(
    swaps: dict[int, int], n_modes: int
) -> NDArray[np.complex128]:
    """
    Calculates the permutation unitary for a given dictionary of swaps across
    the n_modes of a circuit.

    Args:

        swaps (dict) : The dictionary containing the target mode swaps.

        n_modes (int) : The number of modes in the circuit. If this is not the
            number of modes then an incorrect dimension unitary will be
            returned.

    Returns:

        np.ndarray : The determined permutation matrix for the provided set of
            mode swaps.

    """
    if not isinstance(swaps, dict):
        raise TypeError("swaps should be a dictionary object.")

    # Add in missing modes from swap dictionary
    full_swaps = {}
    for m in range(n_modes):
        full_swaps[m] = swaps.get(m, m)
    # Create swap unitary
    permutation = np.zeros((n_modes, n_modes), dtype=complex)
    for i, j in full_swaps.items():
        permutation[j, i] = 1

    return permutation



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.typing import NDArray

from lightworks.sdk.utils.exceptions import DecompositionUnsuccessful
from lightworks.sdk.utils.matrix import check_unitary


def reck_decomposition(
    unitary: NDArray[np.complex128],
) -> tuple[dict[str, float], list[float]]:
    """
    Performs the triangular decomposition procedure for a provided unitary
    matrix.

    Args:

        unitary (np.ndarray) : The unitary matrix on which the decomposition
            should be performed.

    Returns:

        dict : A dictionary which stores the calculated phase shifter settings
            for each element of the interferometer

        list : A list of the residual phases which needs to be applied at the
            output of the system.

    """
    # Check matrix is unitary before decomposition
    if not check_unitary(unitary):
        raise ValueError("Provided matrix determined not to be unitary.")
    n_modes = unitary.shape[0]
    # Dictionary to store calculated phases
    phase_map: dict[str, float] = {}
    # Loop over each element in matrix
    for i in range(0, n_modes - 1, 1):
        for j in range(n_modes - 1 - i):
            # Determine location to null
            loc = n_modes - 1 - i
            # Get elements from unitary
            u_ij = unitary[loc, j]
            u_ij1 = unitary[loc, j + 1]
            # Check if already nulled
            if abs(u_ij) < 1e-20:
                theta, phi = np.pi, 0
            else:
                # Calculate theta and phi
                theta = 2 * np.arctan(abs(u_ij1) / abs(u_ij))
                phi = np.angle(u_ij) - np.angle(u_ij1)
            # Create transformation matrix
            tr_ij = bs_matrix(j, j + 1, theta, phi, n_modes)
            # Null element
            unitary @= np.conj(tr_ij.T)
            phase_map[f"bs_{j + 2 * i}_{j}"] = theta
            phase_map[f"ps_{j + 2 * i}_{j}"] = phi

    # Check matrix has indeed been nulled by code, otherwise raise error
    if not check_null(unitary):
        raise DecompositionUnsuccessful(
            "Unable to successfully perform unitary decomposition procedure."
        )
    # Extract remaining end phases from diagonals of unitary
    end_phases = [np.angle(unitary[i, i]) for i in range(n_modes)]

    return phase_map, end_phases


def bs_matrix(
    mode1: int, mode2: int, theta: float, phi: float, n_modes: int
) -> NDArray[np.complex128]:
    """
    Generates a n_modes X n_modes matrix which implements the beam
    transformation of the unit cell between two modes.
    """
    mat = np.identity(n_modes, dtype=complex)
    gp = 1j * np.exp(1j * theta / 2)
    mat[mode1, mode1] = -np.exp(1j * phi) * np.sin(theta / 2) * gp
    mat[mode1, mode2] = np.cos(theta / 2) * gp
    mat[mode2, mode1] = np.exp(1j * phi) * np.cos(theta / 2) * gp
    mat[mode2, mode2] = np.sin(theta / 2) * gp
    return mat


def check_null(mat: NDArray[np.complex128], precision: float = 1e-10) -> bool:
    """
    A function to check if a provided matrix has been nulled correctly by
    the algorithm.

    Args:

        mat (np.array) : The resultant nulled matrix.

        precision (float, optional) : The precision which the matrix is
            checked according to. If there are large float errors this may
            need to be reduced.

    Returns:

        bool : Indicates if the original unitary was successfully nulled.

    """
    # Loop over each value and ensure it is the expected number to some
    # level of precision
    for i in range(mat.shape[0]):
        for j in range(mat.shape[1]):
            # Check off diagonals are nulled
            if i != j and (
                np.real(mat[i, j] > precision) or np.imag(mat[i, j]) > precision
            ):
                return False  # Return false if any elements aren't
    # Return true if above loop passes
    return True



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import dataclass

from lightworks.emulator.components import Detector, Source
from lightworks.sdk.circuit.photonic_compiler import CompiledPhotonicCircuit
from lightworks.sdk.state import State
from lightworks.sdk.utils.post_selection import PostSelectionType


@dataclass(slots=True)
class TaskData:
    """
    Base class for all task dataclasses. This are used to store all information
    about a task before this is passed to the respective backend for execution.
    """


@dataclass(slots=True)
class AnalyzerTask(TaskData):  # noqa: D101
    circuit: CompiledPhotonicCircuit
    inputs: list[State]
    expected: dict[State, State | list[State]] | None
    post_selection: PostSelectionType | None


@dataclass(slots=True)
class SamplerTask(TaskData):  # noqa: D101
    circuit: CompiledPhotonicCircuit
    input_state: State
    n_samples: int
    source: Source | None
    detector: Detector | None
    post_selection: PostSelectionType | None
    min_detection: int | None
    random_seed: int | None
    sampling_mode: str


@dataclass(slots=True)
class SimulatorTask(TaskData):  # noqa: D101
    circuit: CompiledPhotonicCircuit
    inputs: list[State]
    outputs: list[State] | None



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import UserDict
from math import inf
from numbers import Number
from types import NoneType
from typing import Any, Generic, TypeVar

from lightworks.sdk.utils.exceptions import (
    ParameterBoundsError,
    ParameterDictError,
    ParameterValueError,
)

T = TypeVar("T")


class Parameter(Generic[T]):
    """
    Enables the definition a modifiable parameter that can be used as part of a
    Circuit. It allows for the parameter to be modified after utilisation in a
    Circuit for the adjustment of the functionality with having to redefine the
    entire object. Once created the value of the parameter should be modified
    with the get and set functions.

    Args:

        value (Any) : The value to be assigned to the parameter.

        bounds (list | None, optional) : If specified this allows for
            restrictions to be implemented on the value of each parameter. In
            optimisations this can also used to set parameter bounds if the
            optimisation allows. Bounds should be given as a list in the format
            [min_bound, max_bound]. Note bounds are only supported in the case
            of Numeric values.

        label (str | None, optional) : Used to set an optional label which is
            then shown when using the display circuit method, instead of the
            parameter value.

    """

    def __init__(
        self,
        value: T,
        bounds: list[Number] | None = None,
        label: str | None = None,
    ) -> None:
        # Assign value to attribute
        self.__value = value
        # Process label
        if not isinstance(label, str | NoneType):
            raise TypeError("Label should be a string.")
        self.label = label
        # Process provided bounds
        if not isinstance(bounds, tuple | list | NoneType):
            raise TypeError("Bounds should be a tuple or list.")
        if bounds is not None:
            if len(bounds) != 2:
                raise ValueError(
                    "Bounds should have length 2, containing a min and max "
                    "value for the parameter."
                )
            if not isinstance(value, Number) or isinstance(value, bool):
                raise ParameterBoundsError(
                    "Bounds cannot be set for non-numeric parameters."
                )
            self.min_bound, self.max_bound = bounds[0], bounds[1]
        else:
            self.min_bound, self.max_bound = None, None

    @property
    def min_bound(self) -> Number | None:
        """The lower bound of the parameter value."""
        return self.__min_bound

    @min_bound.setter
    def min_bound(self, value: Any) -> None:
        if value is not None:
            if not is_numeric(self.__value):
                raise ParameterBoundsError(
                    "Bounds cannot be set for non-numeric parameters."
                )
            if not is_numeric(value):
                raise ParameterBoundsError("Bound should be numeric or None.")
            if self.__value < value:
                raise ParameterBoundsError(
                    "Current parameter value is below new minimum bound."
                )
        self.__min_bound = value

    @property
    def max_bound(self) -> Number | None:
        """The upper bound of the parameter value."""
        return self.__max_bound

    @max_bound.setter
    def max_bound(self, value: Any) -> None:
        if value is not None:
            if not is_numeric(self.__value):
                raise ParameterBoundsError(
                    "Bounds cannot be set for non-numeric parameters."
                )
            if not is_numeric(value):
                raise ParameterBoundsError("Bound should be numeric or None.")
            if self.__value > value:
                raise ParameterBoundsError(
                    "Current parameter value is above new maximum bound."
                )
        self.__max_bound = value

    def __str__(self) -> str:
        return str(self.__value)

    def __repr__(self) -> str:
        if isinstance(self.__value, str):
            to_output = f"'{self.__value}'"
        else:
            to_output = str(self.__value)
        if self.label is not None:
            to_output += f", '{self.label}'"
        if self.has_bounds():
            to_output += f", [{self.min_bound}, {self.max_bound}]"
        return f"lightworks.Parameter({to_output})"

    def set(self, value: Any) -> None:
        """Update the current value of the parameter."""
        # Don't allow parameter to be set to non-numeric value if bounds set
        if self.has_bounds() and not is_numeric(value):
            raise ParameterValueError(
                "Parameter cannot be set to non-numeric value when "
                "bounds are assigned to parameter."
            )
        if self.min_bound is not None and value < self.min_bound:
            raise ParameterValueError("Set value is below minimum bound.")
        if self.max_bound is not None and value > self.max_bound:
            raise ParameterValueError("Set value is above maximum bound.")
        self.__value = value

    def get(self) -> T:
        """Returns the current value of the parameter."""
        return self.__value

    def has_bounds(self) -> bool:
        """Checks if the parameter has at least one bound given."""
        return bool(self.min_bound is not None or self.max_bound is not None)


def is_numeric(value: Any) -> bool:
    """General function for checking if a value is numeric."""
    return isinstance(value, Number) and not isinstance(value, bool)


class ParameterDict(UserDict[str, Parameter[Any]]):
    """
    Stores a number of Parameters, using assigned keys to reference each
    Parameter object. This has custom get and set item which allows for the
    parameter object to be retrieved and the parameter value to be changed with
    the [] operator. For example ParameterDict["a"] would return a Parameter
    object and ParameterDict["a"] = 1 would set the value of the Parameter
    associated with the 'a' key to 1.
    """

    def __init__(self, **kwargs: Parameter[Any]) -> None:
        super().__init__({})
        for k, v in kwargs.items():
            if not isinstance(v, Parameter):
                self[k] = Parameter(v)
            else:
                self[k] = v

    @property
    def params(self) -> list[str]:
        """Returns a list of all parameter keys used in the dictionary."""
        return list(self.keys())

    def get_bounds(self) -> dict[str, tuple[Number | float, Number | float]]:
        """
        Retrieves the bounds for all parameters stored in the ParameterDict and
        returns as a dictionary, where the keys match those used to store the
        parameters. If a particular parameter does not have one or both bounds
        set then the bounds will be set to +- inf.
        """
        bounds_dict = {}
        for k, v in super().items():
            minb = v.min_bound if v.min_bound is not None else -inf
            maxb = v.max_bound if v.max_bound is not None else inf
            bounds_dict[k] = (minb, maxb)

        return bounds_dict

    def has_bounds(self) -> bool:
        """
        Checks if any of the parameters stored in the dictionary has a minimum
        and/or maximum bounds associated with it.
        """
        # If any parameters has bounds return True, else return False
        return any(v.has_bounds() for v in self.values())

    def items(self) -> list[tuple[str, Any]]:  # type: ignore[override]
        """Returns pairs of keys and parameter values in a list."""
        return [(k, v.get()) for k, v in super().items()]

    def remove(self, key: Any) -> None:
        """
        Removes the parameter associated with the provided key from the
        ParameterDict.

        Raises:

            KeyError : Raised in cases where the key to remove does not exit in
                the ParameterDict.

        """
        if key not in self:
            raise KeyError("Parameter key not found in ParameterDict.")
        del self[key]

    def __setitem__(self, key: Any, value: Any) -> None:
        """
        Custom set item behaviors to allow for parameter values to be updated
        without overwriting the actual Parameter objects.
        """
        if key in self:
            if isinstance(value, Parameter):
                raise ParameterDictError(
                    "Cannot overwrite existing Parameter with new Parameter "
                    "object."
                )
            super().__getitem__(key).set(value)
        else:
            if not isinstance(value, Parameter):
                raise ParameterDictError(
                    "Values being assigned to new keys should be Parameter "
                    "objects."
                )
            super().__setitem__(key, value)

    def __getitem__(self, key: str) -> Parameter[Any]:
        """
        Implements custom error message for when a parameter key cannot be
        found in the dictionary.
        """
        if key not in self:
            msg = f"Parameter '{key}' not found in ParameterDict."
            raise KeyError(msg)
        return super().__getitem__(key)



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from lightworks import State, StateError


class TestState:
    """Unit tests for State object."""

    def test_get_mode_value(self):
        """Confirms that mode value retrieval works as expected."""
        s = State([1, 2, 3])
        assert s[1] == 2

    def test_state_equality(self):
        """Checks that state equality comparison works as expected."""
        assert State([1, 0, 1, 0]) == State([1, 0, 1, 0])
        assert State([1, 0, 1, 0]) != State([0, 1, 1, 0])

    def test_state_addition(self):
        """Checks that state addition behaviour is as expected."""
        s = State([1, 2, 3]) + State([0, 4, 5])
        assert s == State([1, 2, 3, 0, 4, 5])
        s = State([1, 2, 3]) + State([])
        assert s == State([1, 2, 3])

    def test_merge(self):
        """Checks that state merge works correctly."""
        s = State([1, 2, 3]).merge(State([0, 2, 1]))
        assert s == State([1, 4, 4])

    def test_modification_behavior(self):
        """
        Checks that the correct error is raised if we try to modify the State
        value and checks behaviour if the s attribute of the state is modified.
        """
        s = State([1, 2, 3])
        with pytest.raises(StateError):
            s.s = [1]
        s.s[0] = 2
        assert s == State([1, 2, 3])

    def test_mode_length(self):
        """Check the calculated mode number attribute is set correctly."""
        assert len(State([1, 2, 0, 1, 0])) == 5
        assert len(State([0, 0, 0, 0])) == 4
        assert len(State([])) == 0

    def test_photon_number(self):
        """Checks calculated photon number value is correct."""
        assert State([1, 0, 2, 1, 1]).n_photons == 5
        assert State([0, 0, 0, 0, 0]).n_photons == 0
        assert State([]).n_photons == 0

    @pytest.mark.parametrize(
        "state", [[1, 0, 0, 0], [1, 1, 1, 1], [1, 0, 2, 0], [0, 0, 0, 0]]
    )
    def test_pass_validation(self, state):
        """
        Checks that the _validate method of state does not raise an exception
        when an valid configuration is provided.
        """
        s = State(state)
        s._validate()

    @pytest.mark.parametrize(
        "state", [[1, 0, -1, 0], [1, 1, -1, 0], [1, 0, 0.5, 0], [1.0, 0, 0, 0]]
    )
    def test_fail_validation(self, state):
        """
        Checks that the _validate method of state raises an exception when an
        invalid configuration is provided.
        """
        s = State(state)
        with pytest.raises((ValueError, TypeError)):
            s._validate()



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains a variety of single qubit components which act across a pair of
adjacent dual-rail encoded modes, assuming 0 is the first mode and 1 is the
second mode.
"""

from typing import Any

import numpy as np
import sympy as sp

from lightworks.sdk.circuit import Parameter, Unitary
from lightworks.sdk.utils.param_unitary import ParameterizedUnitary


class I(Unitary):  # noqa: E742
    """
    Implements the identity gate across a pair of modes corresponding to a
    dual-rail encoded qubit.
    """

    def __init__(self) -> None:
        unitary = np.array([[1, 0], [0, 1]])
        super().__init__(unitary, "I")


class H(Unitary):
    """
    Implements a Hadamard across a pair of modes corresponding to a dual-rail
    encoded qubit.
    """

    def __init__(self) -> None:
        unitary = np.array([[1, 1], [1, -1]], dtype=np.complex128) / 2**0.5
        super().__init__(unitary, "H")


class X(Unitary):
    """
    Implements an X gate across a pair of modes corresponding to a dual-rail
    encoded qubit.
    """

    def __init__(self) -> None:
        unitary = np.array([[0, 1], [1, 0]])
        super().__init__(unitary, "X")


class Y(Unitary):
    """
    Implements a Y gate across a pair of modes corresponding to a dual-rail
    encoded qubit.
    """

    def __init__(self) -> None:
        unitary = np.array([[0, -1j], [1j, 0]])
        super().__init__(unitary, "Y")


class Z(Unitary):
    """
    Implements a Z gate across a pair of modes corresponding to a dual-rail
    encoded qubit.
    """

    def __init__(self) -> None:
        unitary = np.array([[1, 0], [0, -1]])
        super().__init__(unitary, "Z")


class S(Unitary):
    """
    Implements an S gate across a pair of modes corresponding to a dual-rail
    encoded qubit.
    """

    def __init__(self) -> None:
        unitary = np.array([[1, 0], [0, 1j]])
        super().__init__(unitary, "S")


class Sadj(Unitary):
    """
    Implements the adjoint S gate across a pair of modes corresponding to a
    dual-rail encoded qubit.
    """

    def __init__(self) -> None:
        unitary = np.array([[1, 0], [0, -1j]])
        super().__init__(unitary, "S_adj")


class T(Unitary):
    """
    Implements a T gate across a pair of modes corresponding to a dual-rail
    encoded qubit.
    """

    def __init__(self) -> None:
        unitary = np.array([[1, 0], [0, np.exp(1j * np.pi / 4)]])
        super().__init__(unitary, "T")


class Tadj(Unitary):
    """
    Implements the adjoint T gate across a pair of modes corresponding to a
    dual-rail encoded qubit.
    """

    def __init__(self) -> None:
        unitary = np.array([[1, 0], [0, np.exp(-1j * np.pi / 4)]])
        super().__init__(unitary, "T_adj")


class SX(Unitary):
    """
    Implements a SX gate across a pair of modes corresponding to a dual-rail
    encoded qubit.
    """

    def __init__(self) -> None:
        unitary = 0.5 * np.array(
            [[1 + 1j, 1 - 1j], [1 - 1j, 1 + 1j]], dtype=np.complex128
        )
        super().__init__(unitary, "SX")


class P(Unitary):
    """
    Implements a phase gate with phase theta across a pair of modes
    corresponding to a dual-rail encoded qubit.

    Args:

        theta (float | Parameter) : The phase angle implemented on the qubit.

    """

    def __init__(self, theta: float | Parameter[float]) -> None:
        name = get_name(theta)
        th = sp.Symbol(name)
        unitary = sp.Matrix(
            [
                [1, 0],
                [0, sp.exp(1j * th)],
            ]
        )
        if isinstance(theta, Parameter):
            super().__init__(
                ParameterizedUnitary(unitary, {name: theta}), f"P({name})"
            )
        else:
            super().__init__(
                np.array(unitary.evalf(subs={name: theta}), dtype=complex),
                f"P({round(theta, 3)})",
            )


class Rx(Unitary):
    """
    Implements a Rx gate with angle theta across a pair of modes corresponding
    to a dual-rail encoded qubit.

    Args:

        theta (float | Parameter) : The rotation angle of the gate.

    """

    def __init__(self, theta: float | Parameter[float]) -> None:
        name = get_name(theta)
        th = sp.Symbol(name)
        unitary = sp.Matrix(
            [
                [sp.cos(th / 2), -1j * sp.sin(th / 2)],
                [-1j * sp.sin(th / 2), sp.cos(th / 2)],
            ]
        )
        if isinstance(theta, Parameter):
            super().__init__(
                ParameterizedUnitary(unitary, {name: theta}), f"Rx({name})"
            )
        else:
            super().__init__(
                np.array(unitary.evalf(subs={name: theta}), dtype=complex),
                f"Rx({round(theta, 3)})",
            )


class Ry(Unitary):
    """
    Implements a Ry gate with angle theta across a pair of modes corresponding
    to a dual-rail encoded qubit.

    Args:

        theta (float | Parameter) : The rotation angle of the gate.

    """

    def __init__(self, theta: float | Parameter[float]) -> None:
        name = get_name(theta)
        th = sp.Symbol(name)
        unitary = sp.Matrix(
            [
                [sp.cos(th / 2), -sp.sin(th / 2)],
                [sp.sin(th / 2), sp.cos(th / 2)],
            ]
        )
        if isinstance(theta, Parameter):
            super().__init__(
                ParameterizedUnitary(unitary, {name: theta}), f"Ry({name})"
            )
        else:
            super().__init__(
                np.array(unitary.evalf(subs={name: theta}), dtype=complex),
                f"Ry({round(theta, 3)})",
            )


class Rz(Unitary):
    """
    Implements a Rz gate with angle theta across a pair of modes corresponding
    to a dual-rail encoded qubit.

    Args:

        theta (float | Parameter) : The rotation angle of the gate.

    """

    def __init__(self, theta: float | Parameter[float]) -> None:
        name = get_name(theta)
        th = sp.Symbol(name)
        unitary = sp.Matrix(
            [
                [sp.exp(-1j * th / 2), 0],
                [0, sp.exp(1j * th / 2)],
            ]
        )
        if isinstance(theta, Parameter):
            super().__init__(
                ParameterizedUnitary(unitary, {name: theta}), f"Rz({name})"
            )
        else:
            super().__init__(
                np.array(unitary.evalf(subs={name: theta}), dtype=complex),
                f"Rz({round(theta, 3)})",
            )


def get_name(param: Any) -> str:
    """
    Returns the name of the parameter if it has a label, otherwise returns
    default theta.
    """
    return (
        param.label
        if isinstance(param, Parameter) and param.label is not None
        else "\u03b8"
    )



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from lightworks.sdk.circuit.parameters import Parameter


def process_parameter_value(value: Any, show_parameter_value: bool) -> Any:
    """
    Takes a provided value and will convert into a label or value if it is
    provided as a parameter.
    """
    if not isinstance(value, Parameter):
        return value
    if show_parameter_value:
        return value.get()
    return value.label if value.label is not None else value.get()



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .distribution import Distribution
from .utils import is_number


class Constant(Distribution):
    """
    Implements returns of a singular constant value for the assigned quantity.

    Args:

        value (float) : The constant value to use.

    """

    def __init__(self, value: float) -> None:
        is_number(value)
        self._value = value

    def __str__(self) -> str:
        return f"Constant({self._value})"

    def value(self) -> float:
        """Returns set constant value."""
        return self._value

    def set_random_seed(self, seed: int | None) -> None:
        """Used for setting the random seed for the model."""



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Lightworks Qubit
================

Enables the implementation of the gate-based paradigm of quantum computing
within linear optic photonic circuits. There is a number of gates, both single
and multi-qubit, and a set of converters for performing conversion from other
quantum programming languages.
"""

from .converter.qiskit_convert import qiskit_converter
from .gates.single_qubit_gates import (
    SX,
    H,
    I,
    P,
    Rx,
    Ry,
    Rz,
    S,
    Sadj,
    T,
    Tadj,
    X,
    Y,
    Z,
)
from .gates.three_qubit_gates import CCNOT, CCZ
from .gates.two_qubit_gates import CNOT, CZ, SWAP, CNOT_Heralded, CZ_Heralded

__all__ = [
    "CCNOT",
    "CCZ",
    "CNOT",
    "CZ",
    "SWAP",
    "SX",
    "CNOT_Heralded",
    "CZ_Heralded",
    "H",
    "I",
    "P",
    "Rx",
    "Ry",
    "Rz",
    "S",
    "Sadj",
    "T",
    "Tadj",
    "X",
    "Y",
    "Z",
    "qiskit_converter",
]



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import randint, random, seed

import numpy as np
import pytest
import sympy as sp

from lightworks import (
    Parameter,
    PostSelection,
    PostSelectionFunction,
    State,
    convert,
    random_permutation,
    random_unitary,
    settings,
)
from lightworks.sdk.circuit.photonic_circuit_utils import (
    add_mode_to_unitary,
    check_loss,
)
from lightworks.sdk.utils.heralding import add_heralds_to_state
from lightworks.sdk.utils.matrix import check_unitary
from lightworks.sdk.utils.param_unitary import ParameterizedUnitary
from lightworks.sdk.utils.permutation import permutation_mat_from_swaps_dict
from lightworks.sdk.utils.post_selection import (
    DefaultPostSelection,
    PostSelectionType,
    Rule,
)
from lightworks.sdk.utils.random import process_random_seed


class TestUtils:
    """
    Unit tests to check functionality of various utilities included with the
    SDK.
    """

    def test_random_unitary(self):
        """
        Checks that when given a seed the random_unitary function always
        produces the same result. If this is not the case it would break many
        of the other unit tests.
        """
        unitary = random_unitary(4, seed=111)
        # Check one diagonal element and two off-diagonals
        assert unitary[0, 0] == pytest.approx(
            -0.49007982458868 + 0.212658840316704j, 1e-8
        )
        assert unitary[1, 2] == pytest.approx(
            -0.3483593186025 - 0.683182137239902j, 1e-8
        )
        assert unitary[3, 2] == pytest.approx(
            0.12574265147702 - 0.1257183128681681j, 1e-8
        )

    def test_random_permutation(self):
        """
        Checks that random permutation consistently returns the same results.
        """
        unitary = random_permutation(4, seed=44)
        # Check one diagonal element and two off-diagonals
        assert unitary[0, 0] == 0j
        assert unitary[0, 1] == 1 + 0j
        assert unitary[3, 0] == 0j

    def test_check_unitary(self):
        """Confirm that the check unitary function behaves as expected."""
        # Check both random unitary and identity matrix
        assert check_unitary(random_unitary(8))
        assert check_unitary(np.identity(8))
        assert check_unitary(np.identity(8, dtype=complex))

    def test_swaps_to_permutations(self):
        """
        Checks that conversion from swaps dict to permutation matrix works as
        expected.
        """
        swaps = {0: 2, 2: 3, 3: 1, 1: 0}
        unitary = permutation_mat_from_swaps_dict(swaps, 4)
        assert abs(unitary[2, 0]) ** 2 == 1
        assert abs(unitary[3, 1]) ** 2 == 0
        assert abs(unitary[3, 2]) ** 2 == 1

    def test_db_loss_to_decimal_conv(self):
        """Test conversion from db loss to a decimal loss value."""
        r = 1 - convert.db_loss_to_decimal(0.5)
        assert r == pytest.approx(0.8912509381337456, 1e-8)

    def test_decimal_to_db_loss_conv(self):
        """
        Tests conversion between a decimal loss and db loss value.
        """
        r = convert.decimal_to_db_loss(0.25)
        assert r == pytest.approx(1.2493873660829995, 1e-8)

    def test_seeded_random(self):
        """
        Checks that the result from the python random module remains consistent
        when using the same seed. If this changes then it could result in other
        unit tests failing.
        """
        seed(999)
        assert random() == pytest.approx(0.7813468849570298, 1e-8)

    @pytest.mark.parametrize("mode", [0, 3, 5, 6])
    def test_add_mode_to_unitary_value(self, mode):
        """
        Checks that add_mode_to_unitary function works correctly for a variety
        of positions, producing a value of 1 in the expected position.
        """
        unitary = random_unitary(6)
        new_unitary = add_mode_to_unitary(unitary, mode)
        # Check diagonal value on new mode
        assert new_unitary[mode, mode] == 1.0
        # Also confirm one off-diagonal value
        assert new_unitary[mode, mode - 1] == 0.0

    @pytest.mark.parametrize("mode", [0, 3, 5, 6])
    def test_add_mode_to_unitary_diagonal(self, mode):
        """
        Checks that add_mode_to_unitary function works correctly for a variety
        of positions.
        """
        unitary = random_unitary(6)
        new_unitary = add_mode_to_unitary(unitary, mode)
        # Confirm preservation of unitary on diagonal
        assert (new_unitary[:mode, :mode] == unitary[:mode, :mode]).all()
        assert (
            new_unitary[mode + 1 :, mode + 1 :] == unitary[mode:, mode:]
        ).all()

    @pytest.mark.parametrize("mode", [0, 3, 5, 6])
    def test_add_mode_to_unitary_off_diagonal(self, mode):
        """
        Checks that add_mode_to_unitary function works correctly for a variety
        of positions.
        """
        unitary = random_unitary(6)
        new_unitary = add_mode_to_unitary(unitary, mode)
        # Confirm preservation on unitary off diagonal
        assert (new_unitary[mode + 1 :, :mode] == unitary[mode:, :mode]).all()
        assert (new_unitary[:mode, mode + 1 :] == unitary[:mode, mode:]).all()

    def test_add_mode_to_parameterized_unitary_return(self):
        """
        Checks that add_mode_to_unitary function returns a Parameterized unitary
        when one is supplied.
        """
        unitary = ParameterizedUnitary(
            sp.Matrix([[6 * i + j for j in range(6)] for i in range(6)]), {}
        )
        new_unitary = add_mode_to_unitary(unitary, 2)
        assert isinstance(new_unitary, ParameterizedUnitary)

    @pytest.mark.parametrize("mode", [0, 3, 5, 6])
    def test_add_mode_to_parameterized_unitary_value(self, mode):
        """
        Checks that add_mode_to_unitary function works correctly for a variety
        of positions, producing a value of 1 in the expected position.
        """
        unitary = ParameterizedUnitary(
            sp.Matrix([[6 * i + j for j in range(6)] for i in range(6)]), {}
        )
        new_unitary = add_mode_to_unitary(unitary, mode)
        # Check diagonal value on new mode
        assert new_unitary._unitary[mode, mode] == 1.0
        # Also confirm one off-diagonal value
        assert new_unitary._unitary[mode, mode - 1] == 0.0

    @pytest.mark.parametrize("mode", [0, 3, 5, 6])
    def test_add_mode_to_parameterized_unitary_diagonal(self, mode):
        """
        Checks that add_mode_to_unitary function works correctly for a variety
        of positions.
        """
        unitary = ParameterizedUnitary(
            sp.Matrix([[6 * i + j for j in range(6)] for i in range(6)]), {}
        )
        new_unitary = add_mode_to_unitary(unitary, mode)
        # Confirm preservation of unitary on diagonal
        assert (
            new_unitary._unitary[:mode, :mode] == unitary._unitary[:mode, :mode]
        )
        assert (
            new_unitary._unitary[mode + 1 :, mode + 1 :]
            == unitary._unitary[mode:, mode:]
        )

    @pytest.mark.parametrize("mode", [0, 3, 5, 6])
    def test_add_mode_to_parameterized_unitary_off_diagonal(self, mode):
        """
        Checks that add_mode_to_unitary function works correctly for a variety
        of positions.
        """
        unitary = ParameterizedUnitary(
            sp.Matrix([[6 * i + j for j in range(6)] for i in range(6)]), {}
        )
        new_unitary = add_mode_to_unitary(unitary, mode)
        # Confirm preservation on unitary off diagonal
        assert (
            new_unitary._unitary[mode + 1 :, :mode]
            == unitary._unitary[mode:, :mode]
        )
        assert (
            new_unitary._unitary[:mode, mode + 1 :]
            == unitary._unitary[:mode, mode:]
        )

    def test_add_heralds_to_state(self):
        """
        Tests that add heralds to state generates the correct mode structure.
        """
        s = [1, 2, 3, 4, 5]
        heralds = {1: 6, 6: 7}
        s_new = add_heralds_to_state(s, heralds)
        assert s_new == [1, 6, 2, 3, 4, 5, 7]

    def test_add_heralds_to_state_unordered(self):
        """
        Tests that add heralds to state generates the correct mode structure,
        when the heralding dict is not ordered.
        """
        s = [1, 2, 3, 4, 5]
        heralds = {6: 7, 1: 6}
        s_new = add_heralds_to_state(s, heralds)
        assert s_new == [1, 6, 2, 3, 4, 5, 7]

    def test_add_heralds_to_state_empty_herald(self):
        """
        Tests that the original state is returned when no heralds are used.
        """
        s = [randint(0, 5) for i in range(10)]
        s_new = add_heralds_to_state(s, {})
        assert s == s_new

    def test_add_heralds_to_state_no_modification(self):
        """
        Checks that add heralds to state does not modify the original state.
        """
        s = [randint(0, 5) for i in range(10)]
        s_copy = list(s)  # Creates copy of list
        add_heralds_to_state(s, {6: 7, 1: 6})
        assert s == s_copy

    def test_add_heralds_to_state_new_object(self):
        """
        Confirms that a new object is still created when no heralds are used
        with a given state.
        """
        s = [randint(0, 5) for i in range(10)]
        s_new = add_heralds_to_state(s, {})
        assert id(s) != id(s_new)

    @pytest.mark.parametrize("value", [1, 3, 2.0, None])
    def test_process_random_seed(self, value):
        """
        Confirms that process_random_seed allows valid values
        """
        process_random_seed(value)

    @pytest.mark.parametrize(
        "value", [0.5, 3.2, "1.1", "seed", [1], (1,), True, False]
    )
    def test_process_random_seed_invalid(self, value):
        """
        Confirms that check_random_seed detects invalid seeds.
        """
        with pytest.raises(TypeError):
            process_random_seed(value)

    @pytest.mark.parametrize(
        "value", [State([1, 0, 0, 1, 1, 0]), [1, 0, 0, 1, 1, 0]]
    )
    def test_valid_dual_rail_to_qubit(self, value):
        """
        Checks conversion from dual-rail to qubit states works correctly for
        valid values.
        """
        conv_state = convert.dual_rail_to_qubit(value)
        assert conv_state == State([0, 1, 0])

    @pytest.mark.parametrize(
        "value", [State([1, 0, 3, 1, 1, 0]), [1, 0, 3, 1, 1, 0]]
    )
    def test_allow_invalid_dual_rail_to_qubit(self, value):
        """
        Checks conversion from dual-rail to qubit states works correctly for
        valid values.
        """
        conv_state = convert.dual_rail_to_qubit(value, allow_invalid=True)
        assert conv_state == State([0, "X", 0])

    @pytest.mark.parametrize(
        "value",
        [
            State([1, 0, 0, 1, 1]),
            State([1, 1, 0, 1, 1, 0]),
            State([0, 0, 0, 1, 1, 0]),
            State([0, 2, 0, 1, 1, 0]),
            [0, 2, 0, 1, 1, 0],
        ],
    )
    def test_invalid_dual_rail_to_qubit(self, value):
        """
        Checks conversion from dual-rail to qubit raises a ValueError for
        incorrect values.
        """
        with pytest.raises(ValueError):
            convert.dual_rail_to_qubit(value)

    @pytest.mark.parametrize("value", [State([0, 1, 0]), [0, 1, 0], "010"])
    def test_valid_qubit_to_dual_rail(self, value):
        """
        Checks conversion from qubit to dual-rail states works correctly for
        valid values.
        """
        conv_state = convert.qubit_to_dual_rail(value)
        assert conv_state == State([1, 0, 0, 1, 1, 0])

    @pytest.mark.parametrize(
        "value", [State([0, 1, 2]), [-1, 1, 0], [0, 1.5, 1], "020"]
    )
    def test_invalid_qubit_to_dual_rail(self, value):
        """
        Checks conversion from qubit to dual-rail states works raises a
        ValueError for incorrect values.
        """
        with pytest.raises(ValueError):
            convert.qubit_to_dual_rail(value)

    def test_state_preservation(self):
        """
        Checks conversion to and from dual rail state to ensure the original
        state is always preserved
        """
        original_state = State([randint(0, 1) for _ in range(10)])
        conv_state = convert.qubit_to_dual_rail(original_state)
        recovered_state = convert.dual_rail_to_qubit(conv_state)
        assert original_state == recovered_state

    def test_threshold_mapping(self):
        """
        Checks threshold mapping produces the correct output.
        """
        original_state = State([1, 0, 2, 3, 0, 0, 1])
        conv_state = convert.threshold_mapping(original_state)
        assert conv_state == State([1, 0, 1, 1, 0, 0, 1])

    def test_threshold_mapping_invert(self):
        """
        Checks threshold mapping produces the correct output when inverting.
        """
        original_state = State([1, 0, 2, 3, 0, 0, 1])
        conv_state = convert.threshold_mapping(original_state, invert=True)
        assert conv_state == State([0, 1, 0, 0, 1, 1, 0])

    def test_parity_mapping(self):
        """
        Checks parity mapping produces the correct output.
        """
        original_state = State([1, 0, 2, 3, 0, 0, 1])
        conv_state = convert.parity_mapping(original_state)
        assert conv_state == State([1, 0, 0, 1, 0, 0, 1])

    def test_parity_mapping_invert(self):
        """
        Checks parity mapping produces the correct output when inverting.
        """
        original_state = State([1, 0, 2, 3, 0, 0, 1])
        conv_state = convert.parity_mapping(original_state, invert=True)
        assert conv_state == State([0, 1, 1, 0, 1, 1, 0])

    @pytest.mark.parametrize(
        "value", [0, 0.0, 0.5, 1, 1.0, Parameter(1), np.float64(0.67)]
    )
    def test_check_loss_valid(self, value):
        """
        Checks an error isn't raised when valid loss are values are provided.
        """
        check_loss(value)

    @pytest.mark.parametrize("value", [True, 0.6j, 1.5, 2, Parameter(2), "1"])
    def test_check_loss_invalid(self, value):
        """
        Checks an error is raised when invalid loss are values are provided.
        """
        with pytest.raises((TypeError, ValueError)):
            check_loss(value)


class TestPostSelection:
    """
    Unit tests for post-selection object.
    """

    def setup_method(self):
        """
        Creates a default state for testing.
        """
        self.test_state = State([1, 0, 1, 0, 1, 0])

    @pytest.mark.parametrize(
        "post_select",
        [
            PostSelection(),
            PostSelectionFunction(lambda s: True),  # noqa: ARG005
            DefaultPostSelection(),
        ],
    )
    def test_is_post_selection_type(self, post_select):
        """
        Checks all post-selection objects are child classes of the
        PostSelectionType.
        """
        assert isinstance(post_select, PostSelectionType)

    @pytest.mark.parametrize(
        ("modes", "photons"),
        [
            (0, 1),
            (0.0, 1.0),
            ((0,), (1,)),
            ((0.0,), (1.0,)),
            ((1, 0), 1),
            (0, (0, 1)),
            ((2, 4), (1, 2)),
        ],
    )
    def test_post_selection(self, modes, photons):
        """
        Tests a range of configuration of the test state and checks they all
        return True.
        """
        p = PostSelection()
        p.add(modes, photons)
        assert p.validate(self.test_state)

    @pytest.mark.parametrize(
        ("modes", "photons"),
        [(0, 0), ((0,), (0,)), ((1, 0), 2), (1, (1, 2)), ((2, 3), (0, 2))],
    )
    def test_post_selection_invalid(self, modes, photons):
        """
        Tests a range of configuration of the test state and checks they all
        return False.
        """
        p = PostSelection()
        p.add(modes, photons)
        assert not p.validate(self.test_state)

    def test_post_selection_duplicate(self):
        """
        Checks that by default post-selection does not allow for two rules on
        the same mode.
        """
        p = PostSelection()
        p.add(1, 1)
        with pytest.raises(ValueError):
            p.add(1, 2)

    def test_post_selection_duplicate_allowed(self):
        """
        Checks that post-selection allows for two rules on the same mode when
        multi_rules specified.
        """
        p = PostSelection(multi_rules=True)
        p.add(1, 1)
        p.add(1, 2)

    def test_post_selection_modes(self):
        """
        Checks that post-selection modes attribute is able to correctly return
        the correct values.
        """
        p = PostSelection()
        p.add(1, 1)
        assert p.modes == [1]
        p.add((2, 3), 1)
        assert p.modes == [1, 2, 3]
        p.add((4,), 1)
        assert p.modes == [1, 2, 3, 4]

    def test_post_selection_duplicates(self):
        """
        Checks that post-selection modes attribute is able to correctly return
        the correct values when a mode is used more than once
        """
        p = PostSelection(multi_rules=True)
        p.add(1, 1)
        assert p.modes == [1]
        p.add((2, 1), 1)
        assert p.modes == [1, 2]
        p.add((2,), 1)
        assert p.modes == [1, 2]

    @pytest.mark.parametrize(
        ("modes", "photons"),
        [
            (1, 1),
            ((1,), (1,)),
            ((1, 2), (1,)),
            ((1,), (1, 2)),
            ((2, 3), (1, 2)),
        ],
    )
    def test_post_selection_rules(self, modes, photons):
        """
        Checks that post-selection rules content is correct.
        """
        p = PostSelection()
        p.add(modes, photons)
        rules = p.rules
        if not isinstance(modes, tuple):
            modes = (modes,)
        if not isinstance(photons, tuple):
            photons = (photons,)
        assert rules[0].modes == modes
        assert rules[0].n_photons == photons

    def test_post_selection_no_rules(self):
        """
        Confirms post-selection object still functions when no rules have been
        provided.
        """
        p = PostSelection()
        assert p.validate(self.test_state)

    def test_post_selection_rules_length(self):
        """
        Confirms that rules property returns a list of the correct length.
        """
        p = PostSelection(multi_rules=True)
        p.add((0, 1), 2)
        p.add((0, 1), 2)
        p.add((2, 3), 1)
        assert len(p.rules) == 3

    @pytest.mark.parametrize("value", [-1, 0.5, ([1, 2],), "1.1"])
    def test_invalid_mode_values(self, value):
        """
        Confirms an error is raised if an invalid mode value is provided.
        """
        p = PostSelection()
        with pytest.raises((ValueError, TypeError)):
            p.add(value, 1)

    @pytest.mark.parametrize("value", [-1, 0.5, ([1, 2],), "1.1"])
    def test_invalid_photon_values(self, value):
        """
        Confirms an error is raised if an invalid photon value is provided.
        """
        p = PostSelection()
        with pytest.raises((ValueError, TypeError)):
            p.add(1, value)

    def test_default_post_selection_always_true(self):
        """
        Checks that default post-selection always returns true by sampling a
        large number of values.
        """
        ps = DefaultPostSelection()
        assert all(ps.validate(self.test_state) is True for _i in range(1000))

    @pytest.mark.parametrize("value", ["not_function", 1, [1, 2, 3]])
    def test_ps_function_enforced(self, value):
        """
        Checks an error is raised if a non-function type is passed to
        PostSelectionFunction.
        """
        with pytest.raises(TypeError):
            PostSelectionFunction(value)

    def test_ps_function_equivalance(self):
        """
        Checks that a function assigned to PostSelectionFunction will always
        return the same value as the function itself.
        """
        func = lambda s: s[1] + s[2] == 1 and s[3] == 0  # noqa: E731
        ps = PostSelectionFunction(func)
        # Check against known state
        assert func(self.test_state) == ps.validate(self.test_state)
        # Then randomly check for 100 states
        states = [[randint(0, 1) for _j in range(6)] for _i in range(100)]
        assert all(func(s) == ps.validate(s) for s in states)

    def test_rule(self):
        """
        Checks a rule can be created which implements a post-selection rule on a
        set of modes/photons.
        """
        Rule((1, 2), (3, 4))

    @pytest.mark.parametrize(
        ("modes", "photons"),
        [((0,), (1,)), ((1, 0), (1,)), ((0,), (0, 1)), ((2, 4), (1, 2))],
    )
    def test_rule_valid(self, modes, photons):
        """
        Desc
        """
        r = Rule(modes, photons)
        assert r.validate(self.test_state)

    def test_rule_tuple(self):
        """
        Checks rule as_tuple method returns expected value.
        """
        r = Rule((1, 2), (3, 4))
        assert r.as_tuple() == ((1, 2), (3, 4))


class TestSettings:
    """
    Set of unit tests for checking the settings module works as expected.
    """

    def test_valid_setting_assignment(self):
        """
        Checks the value of a valid setting can be modified.
        """
        settings.unitary_precision = 1

    def test_valid_setting_retrieval(self):
        """
        Checks the value of a valid setting can be modified and retrieved.
        """
        val = random()
        settings.unitary_precision = val
        assert settings.unitary_precision == val

    def test_invalid_setting_assignment(self):
        """
        Checks attempts to assign an invalid setting will raise an attribute
        error.
        """
        with pytest.raises(AttributeError):
            settings.test_setting = True

    def test_invalid_setting_retrieval(self):
        """
        Checks attempts to retrieve an invalid setting will raise an attribute
        error.
        """
        with pytest.raises(AttributeError):
            settings.test_setting  # noqa: B018

    def test_setting_in_str_repr(self):
        """
        Checks setting names when using str and repr on settings
        """
        assert "sampler_probability_threshold" in str(settings)
        assert "sampler_probability_threshold" in repr(settings)



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from dataclasses import dataclass, field, fields
from numbers import Real
from typing import Any, NoReturn

import numpy as np
from numpy.typing import NDArray

from lightworks.sdk.utils.matrix import check_unitary
from lightworks.sdk.utils.param_unitary import ParameterizedUnitary
from lightworks.sdk.utils.permutation import permutation_mat_from_swaps_dict

from .parameters import Parameter

SERIALIZED_COMPONENT = tuple[str, dict[str, Any]]

# ruff: noqa: D102


@dataclass(slots=True)
class Component(ABC):
    """
    Generic baseclass for all components. Implements a number of useful methods.
    """

    @abstractmethod
    def get_unitary(self, n_modes: int) -> NDArray[np.complex128]:
        """
        Returns a unitary matrix corresponding to the transformation implemented
        by the component with size n_modes.
        """

    @abstractmethod
    def serialize(self) -> SERIALIZED_COMPONENT:
        """
        Creates a serializable tuple of details for the current component.
        """

    def fields(self) -> list[str]:
        """Returns a list of all field from the component dataclass."""
        return [f.name for f in fields(self)]

    def values(self) -> list[Any]:
        """Returns a list of all values from the component dataclass."""
        return [getattr(self, f.name) for f in fields(self)]


@dataclass(slots=True)
class BeamSplitter(Component):
    """
    Configurable beam splitter element between two assigned modes, with support
    for a number of different conventions.
    """

    mode_1: int
    mode_2: int
    reflectivity: float | Parameter[float]
    convention: str

    def __post_init__(self) -> None:
        self.validate()

    def validate(self) -> None:
        """
        Validates that convention and reflectivity values of the beam splitter
        are valid.
        """
        # Validate reflectivity
        if not is_real(self.reflectivity) and not isinstance(
            self.reflectivity, Parameter
        ):
            raise TypeError(
                "Beam splitter reflectivity should a real number or Parameter."
            )
        if (
            not isinstance(self.reflectivity, Parameter)
            and not 0 <= self.reflectivity <= 1
        ):
            raise ValueError("Reflectivity must be in range [0,1].")
        # And check beam splitter convention
        all_convs = ["Rx", "H"]
        if self.convention not in all_convs:
            msg = "Provided beam splitter convention should in ['"
            for conv in all_convs:
                msg += conv + ", "
            msg = msg[:-2] + "']"
            raise ValueError(msg)

    @property
    def _reflectivity(self) -> float:
        if isinstance(self.reflectivity, Parameter):
            return self.reflectivity.get()
        return self.reflectivity

    def get_unitary(self, n_modes: int) -> NDArray[np.complex128]:
        self.validate()
        theta = np.arccos(self._reflectivity**0.5)
        unitary = np.identity(n_modes, dtype=complex)
        if self.convention == "Rx":
            unitary[self.mode_1, self.mode_1] = np.cos(theta)
            unitary[self.mode_1, self.mode_2] = 1j * np.sin(theta)
            unitary[self.mode_2, self.mode_1] = 1j * np.sin(theta)
            unitary[self.mode_2, self.mode_2] = np.cos(theta)
        elif self.convention == "H":
            unitary[self.mode_1, self.mode_1] = np.cos(theta)
            unitary[self.mode_1, self.mode_2] = np.sin(theta)
            unitary[self.mode_2, self.mode_1] = np.sin(theta)
            unitary[self.mode_2, self.mode_2] = -np.cos(theta)
        return unitary

    def serialize(self) -> SERIALIZED_COMPONENT:
        return (
            "BeamSplitter",
            {
                "mode_1": self.mode_1,
                "mode_2": self.mode_2,
                "reflectivity": self._reflectivity,
                "convention": self.convention,
            },
        )


@dataclass(slots=True)
class PhaseShifter(Component):
    """
    Implements a phase shift on the assigned mode.
    """

    mode: int
    phi: float | Parameter[float]

    def __post_init__(self) -> None:
        self.validate()

    def validate(self) -> None:
        """Performs validation of all properties of the phase shifters."""
        if not is_real(self.phi) and not isinstance(self.phi, Parameter):
            raise TypeError(
                "Phase shifter value should a real number or Parameter."
            )

    @property
    def _phi(self) -> float:
        if isinstance(self.phi, Parameter):
            return self.phi.get()
        return self.phi

    def get_unitary(self, n_modes: int) -> NDArray[np.complex128]:
        unitary = np.identity(n_modes, dtype=complex)
        unitary[self.mode, self.mode] = np.exp(1j * self._phi)
        return unitary

    def serialize(self) -> SERIALIZED_COMPONENT:
        return ("PhaseShifter", {"mode": self.mode, "phi": self._phi})


@dataclass(slots=True)
class Loss(Component):
    """
    Induces a loss on the selected circuit mode. This requires creation of
    additional loss modes in the unitary matrix.
    """

    mode: int
    loss: float | Parameter[float]

    def validate(self) -> None:
        """Validates loss value is within allowed range."""
        if not 0 <= self._loss <= 1:
            raise ValueError(
                "Provided loss values should be in the range [0,1]."
            )

    @property
    def _loss(self) -> float:
        if isinstance(self.loss, Parameter):
            return self.loss.get()
        return self.loss

    def get_unitary(self, n_modes: int) -> NDArray[np.complex128]:
        self.validate()
        transmission = 1 - self._loss
        # Assumes loss mode to use is last mode in circuit
        unitary = np.identity(n_modes, dtype=complex)
        unitary[self.mode, self.mode] = transmission**0.5
        unitary[n_modes - 1, n_modes - 1] = transmission**0.5
        unitary[self.mode, n_modes - 1] = (1 - transmission) ** 0.5
        unitary[n_modes - 1, self.mode] = (1 - transmission) ** 0.5
        return unitary

    def serialize(self) -> SERIALIZED_COMPONENT:
        return ("Loss", {"mode": self.mode, "loss": self._loss})


@dataclass(slots=True)
class Barrier(Component):
    """
    Adds a barrier across selected circuit modes.
    """

    modes: list[int]

    def get_unitary(self, n_modes: int) -> NDArray[np.complex128]:
        return np.identity(n_modes, dtype=complex)

    def serialize(self) -> SERIALIZED_COMPONENT:
        return ("Barrier", {"modes": list(self.modes)})


@dataclass(slots=True)
class ModeSwaps(Component):
    """
    Performs ideal swaps between selected modes of the circuit.
    """

    swaps: dict[int, int]

    def __post_init__(self) -> None:
        self.validate()

    def validate(self) -> None:
        """Performs validation of all properties of the mode swaps."""
        # Check swaps are valid
        in_modes = sorted(self.swaps.keys())
        out_modes = sorted(self.swaps.values())
        if in_modes != out_modes:
            raise ValueError(
                "Mode swaps not complete, dictionary keys and values should "
                "contain the same modes."
            )

    def get_unitary(self, n_modes: int) -> NDArray[np.complex128]:
        return permutation_mat_from_swaps_dict(self.swaps, n_modes)

    def serialize(self) -> SERIALIZED_COMPONENT:
        return ("ModeSwaps", {"swaps": self.swaps})


@dataclass(slots=True)
class Group(Component):
    """
    Stores a group of components which have been added to a circuit.
    """

    circuit_spec: list[Component]
    name: str
    mode_1: int
    mode_2: int
    heralds: "HeraldData"

    def get_unitary(self, n_modes: int) -> None:  # type: ignore[override] # noqa: ARG002
        return None

    def serialize(self) -> NoReturn:
        raise RuntimeError(
            "Groups must be unpacked before attempting to serialize"
        )


@dataclass(slots=True)
class UnitaryMatrix(Component):
    """
    Implements a unitary transformation across a subset of circuit modes.
    """

    mode: int
    unitary: NDArray[np.complex128] | ParameterizedUnitary
    label: str

    def __post_init__(self) -> None:
        self.validate()

    def validate(self) -> None:
        """Performs validation of all properties of the unitary matrix."""
        # Check type of supplied unitary
        if isinstance(self.unitary, np.ndarray | list):
            self.unitary = np.array(self.unitary)

            # Ensure unitary is valid
            if not check_unitary(self.unitary):
                raise ValueError("Provided matrix is not unitary.")
        elif not isinstance(self.unitary, ParameterizedUnitary):
            raise TypeError("Unitary should be a numpy array or list.")

        # Also check label is a string
        if not isinstance(self.label, str):
            raise TypeError("Label for unitary should be a string.")

    def get_unitary(self, n_modes: int) -> NDArray[np.complex128]:
        if isinstance(self.unitary, ParameterizedUnitary):
            sub_unitary = self.unitary.unitary
            if not check_unitary(sub_unitary):
                raise ValueError("Provided matrix is not unitary.")
        else:
            sub_unitary = self.unitary
        unitary = np.identity(n_modes, dtype=complex)
        nm = self.unitary.shape[0]
        unitary[self.mode : self.mode + nm, self.mode : self.mode + nm] = (
            sub_unitary[:, :]
        )
        return unitary

    def serialize(self) -> SERIALIZED_COMPONENT:
        if isinstance(self.unitary, ParameterizedUnitary):
            unitary = self.unitary.unitary
            if not check_unitary(unitary):
                raise ValueError("Provided matrix is not unitary.")
        else:
            unitary = self.unitary
        return (
            "UnitaryMatrix",
            {
                "mode": self.mode,
                "unitary": unitary.astype(str).tolist(),
                "label": self.label,
            },
        )


@dataclass(slots=True, frozen=True)
class HeraldData:
    """Stores the heralding data on a circuit"""

    input: dict[int, int] = field(default_factory=dict)
    output: dict[int, int] = field(default_factory=dict)


def is_real(value: Any) -> bool:
    """General function for checking if a value is a real number."""
    return isinstance(value, Real) and not isinstance(value, bool)



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

from .data import TaskData


class Task(ABC):
    """
    Base class for all tasks, which requires implementation of the run method.
    """

    @abstractmethod
    def _generate_task(self) -> TaskData:
        """
        Generates the data required to run the current task.
        """



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lightworks.sdk.circuit import PhotonicCircuit


class _Tomography:
    """
    Base class for all tomography experiments, which contains common logic
    shared across them.

    Args:

        n_qubits (int) : The number of qubits that will be used as part of the
            tomography.

        base_circuit (PhotonicCircuit) : An initial circuit which can be
            modified for performing tomography. It is required that the number
            of circuit input modes equals 2 * the number of qubits.

    """

    def __init__(
        self,
        n_qubits: int,
        base_circuit: PhotonicCircuit,
    ) -> None:
        # Type check inputs
        if not isinstance(n_qubits, int) or isinstance(n_qubits, bool):
            raise TypeError("Number of qubits should be an integer.")
        if not isinstance(base_circuit, PhotonicCircuit):
            raise TypeError("Base circuit should be a circuit object.")

        if 2 * n_qubits != base_circuit.input_modes:
            msg = (
                "Number of circuit input modes does not match the amount "
                "required for the specified number of qubits, expected "
                f"{2 * n_qubits}."
            )
            raise ValueError(msg)

        self._n_qubits = n_qubits
        self._base_circuit = base_circuit

    @property
    def base_circuit(self) -> PhotonicCircuit:
        """
        The base circuit which the tomography experiment will be performed on.
        """
        return self._base_circuit

    @property
    def n_qubits(self) -> int:
        """
        The number of qubits within the system.
        """
        return self._n_qubits



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Class to simulate detector response when sampling, including detector
efficiency, dark counts and non-photon number resolving detectors.
"""

from numbers import Number
from random import random, seed

from lightworks.sdk.state import State


class Detector:
    """
    Creates a detector which can be used to model the presence of imperfect
    detection properties when added to the Sampler.

    Args:

        efficiency (float, optional) : Set the per-channel efficiency of the
            detectors.

        p_dark (float, optional) : The probability of dark counts per detector,
            note that this will depend on the dark count rate and the system
            clock speed.

        photon_counting (bool, optional) : Set to True to use photon number
            resolving detectors and False to use threshold detection.

    """

    def __init__(
        self,
        efficiency: float = 1,
        p_dark: float = 0,
        photon_counting: bool = True,
    ) -> None:
        # Assign to attributes
        self.efficiency = efficiency
        self.p_dark = p_dark
        self.photon_counting = photon_counting

    def __str__(self) -> str:
        return (
            f"Detector(efficiency = {self.efficiency}, p_dark = {self.p_dark},"
            f" photon_counting = {self.photon_counting})"
        )

    def __repr__(self) -> str:
        return "lightworks.emulator." + str(self)

    @property
    def efficiency(self) -> float:
        """The per-channel detection efficiency."""
        return self.__efficiency

    @efficiency.setter
    def efficiency(self, value: float) -> None:
        if not isinstance(value, Number) or isinstance(value, bool):
            raise TypeError("efficiency value should be numeric.")
        if not 0 <= value <= 1:
            raise ValueError("Value of efficiency should be in range [0,1].")
        self.__efficiency = value

    @property
    def p_dark(self) -> float:
        """The per-channel dark counts probability."""
        return self.__p_dark

    @p_dark.setter
    def p_dark(self, value: float) -> None:
        if not isinstance(value, Number) or isinstance(value, bool):
            raise TypeError("p_dark value should be numeric.")
        if not 0 <= value <= 1:
            raise ValueError("Value of p_dark should be in range [0,1].")
        self.__p_dark = value

    @property
    def photon_counting(self) -> float:
        """Controls whether the detectors are photon number resolving."""
        return self.__photon_counting

    @photon_counting.setter
    def photon_counting(self, value: float) -> None:
        if not isinstance(value, bool):
            raise TypeError("photon_counting should be a boolean.")
        self.__photon_counting = value

    def _get_output(self, in_state: State) -> State:
        """
        Sample an output state from the provided input.

        Args:

            in_state (State) : The input state to the detection module.

        Returns:

            State: The processed output state.

        """
        # If detectors are perfect then just return input
        if self.efficiency == 1 and self.p_dark == 0 and self.photon_counting:
            return in_state
        # Convert state to list
        output = list(in_state)
        # Account for efficiency
        if self.efficiency < 1:
            for mode, n in enumerate(in_state):
                for _i in range(n):
                    if random() > self.efficiency:
                        output[mode] -= 1
        # Then include dark counts
        if self.p_dark > 0:
            for mode in range(len(in_state)):
                if random() < self.p_dark:
                    output[mode] += 1
        # Also account for non-photon counting detectors
        if not self.photon_counting:
            output = [1 if count >= 1 else 0 for count in output]

        return State(output)

    def _set_random_seed(self, r_seed: float | None) -> None:
        """
        Set the random seed for the detector to produce repeatable results.
        """
        seed(r_seed)



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

import numpy as np
from numpy.typing import NDArray

from lightworks.emulator.utils.sim import check_photon_numbers
from lightworks.emulator.utils.state import fock_basis
from lightworks.sdk.results import SimulationResult
from lightworks.sdk.state import State
from lightworks.sdk.tasks import SimulatorTask
from lightworks.sdk.utils.heralding import add_heralds_to_state

from .runner import RunnerABC


class SimulatorRunner(RunnerABC):
    """
    Calculates the probability amplitudes between a set of inputs and outputs
    from a given circuit.

    Args:

        data (SimulatorTask) : The task which is to be executed.

        amplitude_function (Callable) : Function for calculating probability
            amplitudes between an input and output for a given unitary.

    """

    def __init__(
        self,
        data: SimulatorTask,
        amplitude_function: Callable[
            [NDArray[np.complex128], list[int], list[int]], complex
        ],
    ) -> None:
        self.data = data
        self.func = amplitude_function

    def run(self) -> SimulationResult:
        """
        Runs the simulation task.

        Returns:

            SimulationResult : A dictionary containing the calculated
                probability amplitudes, where the first index of the array
                corresponds to the input state, as well as the input and output
                state used to create the array.

        """
        in_heralds = self.data.circuit.heralds.input
        out_heralds = self.data.circuit.heralds.output
        in_heralds_n = sum(in_heralds.values())
        out_heralds_n = sum(out_heralds.values())
        target_n = self.data.inputs[0].n_photons + in_heralds_n
        if self.data.outputs is None:
            check_photon_numbers(self.data.inputs, target_n - in_heralds_n)
            outputs = [
                State(s)
                for s in fock_basis(
                    self.data.circuit.input_modes, target_n - out_heralds_n
                )
            ]
        else:
            check_photon_numbers(self.data.inputs, target_n - in_heralds_n)
            check_photon_numbers(self.data.outputs, target_n - out_heralds_n)
            outputs = self.data.outputs
        # Pre-add output values to avoid doing this many times
        full_outputs = [
            add_heralds_to_state(outs, out_heralds)
            + [0] * self.data.circuit.loss_modes
            for outs in outputs
        ]
        # Calculate permanent for the given inputs and outputs and return
        # values
        amplitudes = np.zeros(
            (len(self.data.inputs), len(outputs)), dtype=complex
        )
        for i, ins in enumerate(self.data.inputs):
            in_state = add_heralds_to_state(ins, in_heralds)
            in_state += [0] * self.data.circuit.loss_modes
            for j, outs in enumerate(full_outputs):
                amplitudes[i, j] = self.func(
                    self.data.circuit.U_full, in_state, outs
                )
        # Return results and corresponding states as dictionary
        return SimulationResult(
            amplitudes,
            "probability_amplitude",
            inputs=self.data.inputs,
            outputs=outputs,
        )



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from lightworks import random_unitary
from lightworks.tomography import (
    choi_from_unitary,
    density_from_state,
    process_fidelity,
    state_fidelity,
)
from lightworks.tomography.utils import (
    _check_target_process,
    _get_required_tomo_measurements,
    _get_tomo_measurements,
)

U_CNOT = np.array(
    [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]], dtype=complex
)
U_CCNOT = np.identity(8, dtype=complex)
U_CCNOT[6:, :] = U_CCNOT[7:5:-1, :]
U_CCZ = np.identity(8, dtype=complex)
U_CCZ[7, 7] = -1
# NOTE: Currently an issue with scipy.linalg.sqrtm on certain platforms, which
# causes process fidelity to fail. This can be mitigated by applying a small
# permutation to one of the elements in the unitary. This is unlikely to be an
# issue in real scenarios.
for mat in [U_CNOT, U_CCNOT, U_CCZ]:
    mat[0, 0] = 0.9999 + 0.0001j


class TestUtils:
    """
    Unit tests for utilities of the tomography module.
    """

    @pytest.mark.parametrize(
        "rho",
        [
            [[1, 0], [0, 0]],
            [[0.5, 0.5], [0.5, 0.5]],
            [[0.5, 0.5j], [-0.5j, 0.5]],
        ],
    )
    def test_state_fidelity(self, rho):
        """
        Validate that fidelity value is always 1 when using two identical
        matrices.
        """
        rho = np.array(rho, dtype=complex)
        assert state_fidelity(rho, rho) == pytest.approx(1, 1e-6)

    def test_state_fidelity_dim_mismatch(self):
        """
        Check an error is raised if there is a mismatch in dimensions between
        density matrices.
        """
        with pytest.raises(ValueError, match="dimensions"):
            state_fidelity(random_unitary(3), random_unitary(4))

    @pytest.mark.parametrize(
        "choi",
        [
            choi_from_unitary([[0, 1], [1, 0]]),
            choi_from_unitary([[0, -1j], [1j, 0]]),
            choi_from_unitary([[2**-0.5, 2**-0.5], [2**-0.5, -(2**-0.5)]]),
            choi_from_unitary(U_CNOT),
            choi_from_unitary(U_CCNOT),
            choi_from_unitary(U_CCZ),
        ],
    )
    def test_process_fidelity(self, choi):
        """
        Validate that fidelity value is always 1 when using two identical
        matrices.
        """
        assert process_fidelity(choi, choi) == pytest.approx(1, 1e-3)

    def test_process_fidelity_dim_mismatch(self):
        """
        Check an error is raised if there is a mismatch in dimensions between
        density matrices.
        """
        with pytest.raises(ValueError, match="dimensions"):
            process_fidelity(random_unitary(3), random_unitary(4))

    def test_basic_density_matrix_calc(self):
        """
        Checks the density matrix of the two qubit state |00> is correct.
        """
        rho = density_from_state([1, 0, 0, 0])
        assert (
            rho
            == np.array(
                [[1, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]
            )
        ).all()

    def test_bell_state_density_matrix_calc(self):
        """
        Checks the calculated density matrix for a bell state is correct.
        """
        rho = density_from_state([2**-0.5, 0, 0, 2**-0.5])
        assert (
            rho.round(5)
            == np.array(
                [[0.5, 0, 0, 0.5], [0, 0, 0, 0], [0, 0, 0, 0], [0.5, 0, 0, 0.5]]
            )
        ).all()

    def test_hadamard_choi(self):
        """
        Checks that the calculated choi matrix for the hadamard gate is correct.
        """
        choi = choi_from_unitary([[2**-0.5, 2**-0.5], [2**-0.5, -(2**-0.5)]])
        assert (
            choi.round(5)
            == np.array(
                [
                    [0.5, 0.5, 0.5, -0.5],
                    [0.5, 0.5, 0.5, -0.5],
                    [0.5, 0.5, 0.5, -0.5],
                    [-0.5, -0.5, -0.5, 0.5],
                ]
            )
        ).all()

    @pytest.mark.parametrize("n_qubits", [1, 3])
    def test_number_of_measurements(self, n_qubits):
        """
        Confirms that the number of measurements for a full tomography is
        4^n_qubits.
        """
        assert len(_get_tomo_measurements(n_qubits)) == 4**n_qubits

    @pytest.mark.parametrize("remove", [True, False])
    def test_measurements_remove_trivial(self, remove):
        """
        Checks that the trivial I,I measurement is included when required and
        then removed otherwise
        """
        all_meas = _get_tomo_measurements(2, remove_trivial=remove)
        if not remove:
            assert "I,I" in all_meas
        else:
            assert "I,I" not in all_meas

    @pytest.mark.parametrize("n_qubits", [1, 3])
    def test_number_of_required_measurements(self, n_qubits):
        """
        Confirms that the number of required measurements for a tomography is
        3^n_qubits.
        """
        assert len(_get_required_tomo_measurements(n_qubits)[0]) == 3**n_qubits

    @pytest.mark.parametrize("n_qubits", [1, 3])
    def test_required_measurements_order(self, n_qubits):
        """
        Checks that the required measurements function always returns a sorted
        list for consistency between runs.
        """
        req_meas = _get_required_tomo_measurements(n_qubits)[0]
        assert req_meas == sorted(req_meas)

    def test_check_target_process(self):
        """
        Confirm no error is raised when a valid process matrix is provided to
        _check_target_process.
        """
        _check_target_process(random_unitary(4), 2)

    @pytest.mark.parametrize("value", ["Test", {1: 2, 3: 4}, [[1, 2], [3]]])
    def test_check_target_process_type(self, value):
        """
        Confirms an error is raised when an invalid type is provided to
        _check_target_process.
        """
        with pytest.raises(TypeError):
            _check_target_process(value, 2)

    @pytest.mark.parametrize(
        "value",
        [
            random_unitary(4)[:3, :3],
            random_unitary(4)[:, :3],
            random_unitary(4)[:3, :],
        ],
    )
    def test_check_target_process_dimension(self, value):
        """
        Confirms an error is raised when the target process has incorrect
        "dimension.
        """
        with pytest.raises(ValueError):
            _check_target_process(value, 2)



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lightworks.sdk.circuit import PhotonicCircuit
from lightworks.sdk.state import State

from .exceptions import ModeMismatchError, PhotonNumberError


def validate_states(states: State | list[State], n_modes: int) -> list[State]:
    """
    Performs all required processing/checking a list of states from a task.
    """
    # Convert state to list of States if not provided for single state case
    if isinstance(states, State):
        states = [states]
    # Check each input
    for s in states:
        # Ensure correct type
        if not isinstance(s, State):
            raise TypeError(
                "Assigned states should be a State or list of State objects."
            )
        # Dimension check
        if len(s) != n_modes:
            msg = (
                "One or more states have an incorrect number of modes, correct "
                f"number of modes is {n_modes}."
            )
            raise ModeMismatchError(msg)
        # Also validate state values
        s._validate()
    return states


def check_herald_difference(
    circuit: PhotonicCircuit, input_photons: int
) -> None:
    """
    Validates that the number of output heralds is not too large based on the
    number of inputs to the circuit.
    """
    photon_diff = sum(circuit.heralds.output.values()) - sum(
        circuit.heralds.input.values()
    )
    if photon_diff > input_photons:
        msg = (
            "Number of input photons is smaller than that required based "
            "on the number of heralds at the output. At least "
            f"{photon_diff} input photons are required."
        )
        raise PhotonNumberError(msg)



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import TYPE_CHECKING, Any, overload

from multimethod import multimethod

from lightworks.emulator.utils.exceptions import BackendError
from lightworks.sdk.results import Result
from lightworks.sdk.tasks import Batch, Task

from .permanent import PermanentBackend
from .slos import SLOSBackend

if TYPE_CHECKING:
    from .abc_backend import EmulatorBackend


class Backend:
    """
    Provide central location for selecting and interacting with different
    simulation backends.

    Args:

        backend (str) : A string detailing the backend which is to be used.

    """

    def __init__(self, backend: str) -> None:
        self.backend = backend

    @overload
    def run(self, task: Task) -> Result[Any, Any]: ...

    @overload
    def run(self, task: Batch) -> list[Result[Any, Any]]: ...

    def run(
        self, task: Task | Batch
    ) -> Result[Any, Any] | list[Result[Any, Any]]:
        """
        Runs the provided task on the current backend.

        Args:

            task (Task|Batch) : A task or batch to run.

        Returns:

            Result|list[Result]: A dictionary like results object containing
                details of the calculated values from a task. If a batch is run
                then this will be a list of results in the same order the task
                were added to the batch.

        """
        if not isinstance(task, Task | Batch):
            raise TypeError("Object to run on the backend must be a task.")
        return self._run(task)

    @multimethod
    def _run(self, task: Task) -> Result[Any, Any]:
        if task.__class__.__name__ not in self.__backend.compatible_tasks:
            msg = (
                "Selected backend not compatible with task, supported tasks for"
                " the backend are: "
                f"{', '.join(self.__backend.compatible_tasks)}."
            )
            raise BackendError(msg)
        return self.__backend.run(task)

    @_run.register
    def _run_batch(self, task: Batch) -> list[Result[Any, Any]]:
        return [self._run(t) for t in task]

    @property
    def backend(self) -> str:
        """
        Returns the name of the currently selected backend.
        """
        return self.__backend.name

    @backend.setter
    def backend(self, value: str) -> None:
        backends = {
            "permanent": PermanentBackend,
            "slos": SLOSBackend,
        }
        if value not in backends:
            msg = (
                "Backend name not recognised, valid options are: "
                f"{', '.join(backends.keys())}."
            )
            raise ValueError(msg)
        # Initialise selected backend
        self.__backend: EmulatorBackend = backends[value]()

    def __str__(self) -> str:
        return self.backend

    def __repr__(self) -> str:
        return f"lightworks.emulator.Backend('{self.backend}')"



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy import random

from .distribution import Distribution
from .utils import is_number


class Gaussian(Distribution):
    """
    Returns random values according to a Gaussian distribution with defined
    center and standard deviation. It can also be constrained to be between a
    minimum and maximum value to prevent issues with assigning invalid
    quantities. When a value is outside of the set bounds the distribution will
    be resampled until this is no longer the case. Note: care should be taken
    with setting minimum and maximum values as setting these to be too strict
    could significantly increase the time taken to produce a valid value.

    Args:

        center (float) : The center (mean) of the Gaussian distribution.

        deviation (float) : The standard deviation of the distribution.

        min_value (float | None) : The minimum allowed value for the
            distribution. Defaults to None, which will assign the min value to
            be - infinity.

        max_value (float | None) : The maximum allowed value for the
            distribution. Defaults to None, which will assign the max value to
            be + infinity.

    """

    def __init__(
        self,
        center: float,
        deviation: float,
        min_value: float | None = None,
        max_value: float | None = None,
    ) -> None:
        # Assign min\max to +/i infinity if None
        if min_value is None:
            min_value = -np.inf
        if max_value is None:
            max_value = np.inf
        # Type check all values
        is_number([center, deviation, min_value, max_value])
        # Then check min and max values are valid
        if max_value < min_value:
            raise ValueError("Max value cannot be less than min value.")
        # Then assign to attributes
        self._center = center
        self._deviation = deviation
        self._min_value = min_value
        self._max_value = max_value
        self._rng = random.default_rng()

    def __str__(self) -> str:
        if self._min_value == -np.inf and self._max_value == np.inf:
            contents = f"\u03bc = {self._center}, \u03c3 = {self._deviation}"
        else:
            contents = (
                f"\u03bc = {self._center}, \u03c3 = {self._deviation}, "
                f"bounds = [{self._min_value}, {self._max_value}]"
            )
        return f"Gaussian({contents})"

    def value(self) -> float:
        """Returns random value from the Gaussian distribution."""
        val = self._rng.normal(self._center, self._deviation)
        # Recalculate value until valid.
        while val < self._min_value or val > self._max_value:
            val = self._rng.normal(self._center, self._deviation)
        # Then return
        return val

    def set_random_seed(self, seed: int | None) -> None:
        """Used for setting the random seed for the model."""
        self._rng = random.default_rng(seed)



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains a variety of three qubit components, designed for implementing required
qubit processing functionality in lightworks.
"""

import numpy as np

from lightworks.sdk.circuit import PhotonicCircuit, Unitary

from .single_qubit_gates import H


class CCZ(PhotonicCircuit):
    """
    Post-selected CCZ gate which acts across three dual-rail encoded qubits.
    There is a total of 4 heralded modes also included, each requiring 0
    photons on the input and output. For correct functioning of this gate it
    must be post-selected on the condition that only one photon is measured
    across the two modes used to encode each of the qubits.
    """

    def __init__(self) -> None:
        # Define transformation matrix
        # fmt: off
        u_a = np.array(
            [[3**-0.5, 0, 0, 1j * (2 / 3)**0.5, 0, 0, 0, 0, 0, 0],
             [0, 3**-0.5, 0, 0, 0, 1j * (2 / 3)**0.5, 0, 0, 0, 0],
             [0, 0, - 3**-0.5, 0, 0, 0, 0, 2**-0.5, -1j * 6**-0.5, 0],
             [-1j * (2 / 3)**0.5, 0, 0, -3**-0.5, 0, 0, 0, 0, 0, 0],
             [0, 0, -1j * 2**0.5 / 3, 0, -3**-0.5, 0, 0, -1j * 3**-0.5, 1 / 3,
              0],
             [0, -1j * (2 / 3)**0.5, 0, 0, 0, -3**-0.5, 0, 0, 0, 0],
             [0, 0, 0, 0, 0, 0, 1j * 2**-1.5, 0, 0, 1j * (7 / 8)**0.5],
             [0, 0, -1j * 3**-0.5, 0, 2**-0.5, 0, 0, -1j * 2**-1.5, -24**-0.5,
              0],
             [0, 0, -1j / 3, 0, -6**-0.5, 0, 0, 1j * 24**-0.5, -7 * 72**-0.5,
              0],
             [0, 0, 0, 0, 0, 0, -(7 / 8)**0.5, 0, 0, 2**-1.5]]
        )
        # fmt: on
        # Create unitary component and add heralds on required modes
        unitary = Unitary(u_a)
        unitary.herald(0, 0)
        unitary.herald(1, 0)
        unitary.herald(8, 0)
        unitary.herald(9, 0)

        super().__init__(6)
        self.add(unitary, 0, group=True, name="CCZ")


class CCNOT(PhotonicCircuit):
    """
    Post-selected CCNOT (Toffoli) gate which acts across three dual-rail
    encoded qubits. There is a total of 4 heralded modes also included, each
    requiring 0 photons on the input and output. For correct functioning of
    this gate it must be post-selected on the condition that only one photon is
    measured across the two modes used to encode each of the qubits.

    Args:

        target_qubit (int, optional) : Sets which of the three qubits is used as
            the target qubit for the gate. Should be either 0, 1 or 2.

    """

    def __init__(self, target_qubit: int = 2) -> None:
        if target_qubit not in {0, 1, 2}:
            raise ValueError(
                "target_qubit setting must have a value of either 0, 1 or 3."
            )

        # Create CCNOT from combination of H and CCZ
        circ = PhotonicCircuit(6)
        circ.add(H(), 2 * target_qubit)
        circ.add(CCZ(), 0)
        circ.add(H(), 2 * target_qubit)

        super().__init__(6)

        controls = tuple(i for i in range(3) if i != target_qubit)
        name = f"CCNOT ({controls}, {target_qubit})"
        self.add(circ, 0, group=True, name=name)



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains a collection of different useful functions for operations on matrices.
"""

import numpy as np
from numpy.typing import NDArray

from lightworks.__settings import settings


def check_unitary(
    U: NDArray[np.complex128],  # noqa: N803
    precision: float | None = None,
) -> bool:
    """
    A function to check if a provided matrix is unitary according to a
    certain level of precision. If finds the product of the matrix with its
    hermitian conjugate and then checks it is unitary.

    Args:

        U (np.array) : The NxN matrix which we want to check is unitary.

        precision (float, optional) : The precision which the unitary
            matrix is checked according to. If there are large float errors
            this may need to be reduced.

    Returns:

        bool : A boolean to indicate whether or not the matrix is unitary.

    Raises:

        ValueError : Raised in the event that the matrix is not square as it
            cannot be unitary.

    """
    if precision is None:
        precision = settings.unitary_precision
    if U.shape[0] != U.shape[1]:
        raise ValueError("Unitary matrix must be square.")
    # Find hermitian conjugate and then product
    hc = np.conj(np.transpose(U))
    # Validate close according to tolerance
    return np.allclose(
        hc @ U, np.identity(U.shape[0], dtype=complex), rtol=0, atol=precision
    )
